#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

from qt.core import QAbstractItemView, QAbstractListModel, QIcon, QItemSelectionModel, Qt

from calibre import force_unicode
from calibre.gui2 import error_dialog, gprefs, warning_dialog
from calibre.gui2.preferences import AbortCommit, ConfigWidgetBase, test_widget
from calibre.gui2.preferences.toolbar_ui import Ui_Form
from calibre.startup import connect_lambda
from calibre.utils.icu import primary_sort_key
from calibre.utils.localization import _


def sort_key_for_action(ac):
    q = getattr(ac, 'action_spec', None)
    try:
        q = ac.name if q is None else q[0]
        return primary_sort_key(force_unicode(q))
    except Exception:
        return primary_sort_key('')


class FakeAction:
    def __init__(self, name, gui_name, icon, tooltip=None, dont_add_to=frozenset(), dont_remove_from=frozenset()):
        self.name = name
        self.action_spec = (gui_name, icon, tooltip, None)
        self.dont_remove_from = dont_remove_from
        self.dont_add_to = dont_add_to


class BaseModel(QAbstractListModel):
    _data: list

    def name_to_action(self, name, gui):
        if name == 'Donate':
            return FakeAction(
                'Donate',
                _('Donate'),
                'donate.png',
                tooltip=_('Donate to support the development of calibre'),
                dont_add_to=frozenset(['context-menu', 'context-menu-device', 'searchbar']),
            )
        if name == 'Location Manager':
            return FakeAction(
                'Location Manager',
                _('Location Manager'),
                'reader.png',
                _('Switch between library and device views'),
                dont_add_to=frozenset([
                    'menubar',
                    'toolbar',
                    'toolbar-child',
                    'context-menu',
                    'searchbar',
                    'context-menu-device',
                ]),
            )
        if name is None:
            return FakeAction(
                '--- ' + ('Separator') + ' ---',
                '--- ' + _('Separator') + ' ---',
                None,
                dont_add_to=frozenset(['menubar', 'menubar-device']),
            )
        try:
            return gui.iactions[name]
        except Exception:
            return None

    def rowCount(self, parent=...):
        return len(self._data)

    def data(self, index, role=...):
        row = index.row()
        action = self._data[row].action_spec
        if role == Qt.ItemDataRole.DisplayRole:
            text = action[0]
            text = text.replace('&', '')
            if text == _('%d books'):
                text = _('Choose library')
            return text
        if role == Qt.ItemDataRole.DecorationRole:
            if hasattr(self._data[row], 'qaction'):
                icon = self._data[row].qaction.icon()
                if not icon.isNull():
                    return icon
            ic = action[1]
            if ic is None:
                ic = 'blank.png'
            return QIcon.ic(ic)
        if role == Qt.ItemDataRole.ToolTipRole and action[2] is not None:
            return action[2]
        return None

    def names(self, indexes):
        rows = [i.row() for i in indexes]
        ans = []
        for i in rows:
            n = self._data[i].name
            if n.startswith('---'):
                n = None
            ans.append(n)
        return ans

    def has_action(self, name):
        for a in self._data:
            if a.name == name:
                return True
        return False


class AllModel(BaseModel):
    def __init__(self, key, gui):
        BaseModel.__init__(self)
        self.gprefs_name = 'action-layout-' + key
        current = gprefs[self.gprefs_name]
        self.gui = gui
        self.key = key
        self._data = self.get_all_actions(current)

    def get_all_actions(self, current):
        all = list(self.gui.iactions.keys()) + ['Donate', 'Location Manager']
        all = [x for x in all if x not in current] + [None]
        all = [self.name_to_action(x, self.gui) for x in all]
        all = [x for x in all if self.key not in x.dont_add_to]

        all.sort(key=sort_key_for_action)
        return all

    def add(self, names):
        actions = []
        for name in names:
            if name is None or name.startswith('---'):
                continue
            actions.append(self.name_to_action(name, self.gui))
        self.beginResetModel()
        self._data.extend(actions)
        self._data.sort(key=sort_key_for_action)
        self.endResetModel()

    def remove(self, indices, allowed):
        rows = [i.row() for i in indices]
        remove = set()
        for row in rows:
            ac = self._data[row]
            if ac.name.startswith('---'):
                continue
            if ac.name in allowed:
                remove.add(row)
        ndata = []
        for i, ac in enumerate(self._data):
            if i not in remove:
                ndata.append(ac)
        self.beginResetModel()
        self._data = ndata
        self.endResetModel()

    def restore_defaults(self):
        current = gprefs.defaults[self.gprefs_name]
        self.beginResetModel()
        self._data = self.get_all_actions(current)
        self.endResetModel()


class CurrentModel(BaseModel):
    def __init__(self, key, gui):
        BaseModel.__init__(self)
        self.gprefs_name = 'action-layout-' + key
        current = gprefs[self.gprefs_name]
        self._data = [self.name_to_action(x, gui) for x in current]
        self._data = [x for x in self._data if x is not None]
        self.key = key
        self.gui = gui

    def move_single(self, idx, delta):
        row = idx.row()
        nrow = (row + delta + len(self._data)) % len(self._data)
        if row + delta < 0:
            x = self._data.pop(row)
            self._data.append(x)
        elif row + delta >= len(self._data):
            x = self._data.pop(row)
            self._data.insert(0, x)
        else:
            self._data[row], self._data[nrow] = self._data[nrow], self._data[row]

    def move_many(self, indices, delta):
        rows = [i.row() for i in indices]
        items = [self._data[x.row()] for x in indices]
        self.beginResetModel()
        indices = sorted(indices, key=lambda i: i.row(), reverse=delta > 0)
        for item in items:
            self.move_single(self.index(self._data.index(item)), delta)
        self.endResetModel()
        ans = {}
        for item, row in zip(items, rows):
            ans[row] = self.index(self._data.index(item))
        return ans

    def add(self, names):
        actions = []
        reject = set()
        for name in names:
            ac = self.name_to_action(name, self.gui)
            if self.key in ac.dont_add_to:
                reject.add(ac)
            else:
                actions.append(ac)

        self.beginResetModel()
        self._data.extend(actions)
        self.endResetModel()
        return reject

    def remove(self, indices):
        rows = [i.row() for i in indices]
        remove, rejected = set(), set()
        for row in rows:
            ac = self._data[row]
            if self.key in ac.dont_remove_from:
                rejected.add(ac)
                continue
            remove.add(row)
        ndata = []
        for i, ac in enumerate(self._data):
            if i not in remove:
                ndata.append(ac)
        self.beginResetModel()
        self._data = ndata
        self.endResetModel()
        return rejected

    def commit(self):
        old = gprefs[self.gprefs_name]
        new = []
        for x in self._data:
            n = x.name
            if n.startswith('---'):
                n = None
            new.append(n)
        new = tuple(new)
        if new != old:
            defaults = gprefs.defaults[self.gprefs_name]
            if defaults == new:
                del gprefs[self.gprefs_name]
            else:
                gprefs[self.gprefs_name] = new

    def restore_defaults(self):
        current = gprefs.defaults[self.gprefs_name]
        self.beginResetModel()
        self._data = [self.name_to_action(x, self.gui) for x in current]
        self.endResetModel()


class ConfigWidget(ConfigWidgetBase, Ui_Form):
    LOCATIONS = [
        ('toolbar', _('The main toolbar')),
        ('toolbar-device', _('The main toolbar when a device is connected')),
        ('toolbar-child', _('The optional second toolbar')),
        ('searchbar', _('The buttons on the search bar')),
        ('menubar', _('The menubar')),
        ('menubar-device', _('The menubar when a device is connected')),
        ('context-menu', _('The context menu for the books in the calibre library')),
        ('context-menu-split', _('The context menu for the split book list')),
        ('context-menu-device', _('The context menu for the books on the device')),
        ('context-menu-cover-browser', _('The context menu for the Cover browser')),
    ]

    def genesis(self, gui):
        self.all_actions.doubleClicked.connect(self.add_single_action)
        self.current_actions.doubleClicked.connect(self.remove_single_action)
        self.models = {}
        self.what.addItem(_('Click to choose toolbar or menu to customize'), 'blank')
        for key, text in self.LOCATIONS:
            self.what.addItem(text, key)
            all_model = AllModel(key, gui)
            current_model = CurrentModel(key, gui)
            self.models[key] = (all_model, current_model)
        self.what.setCurrentIndex(0)
        self.what.currentIndexChanged.connect(self.what_changed)
        self.what_changed(0)

        self.add_action_button.clicked.connect(self.add_action)
        self.remove_action_button.clicked.connect(self.remove_action)
        connect_lambda(self.action_up_button.clicked, self, lambda self: self._move_action(-1))
        connect_lambda(self.action_down_button.clicked, self, lambda self: self._move_action(1))
        self.all_actions.setMouseTracking(True)
        self.current_actions.setMouseTracking(True)
        self.all_actions.entered.connect(self.all_entered)
        self.current_actions.entered.connect(self.current_entered)

    def all_entered(self, index):
        all_model = self.all_actions.model()
        assert all_model is not None
        tt = all_model.data(index, Qt.ItemDataRole.ToolTipRole) or ''
        self.help_text.setText(tt)

    def current_entered(self, index):
        cur_model = self.current_actions.model()
        assert cur_model is not None
        tt = cur_model.data(index, Qt.ItemDataRole.ToolTipRole) or ''
        self.help_text.setText(tt)

    def what_changed(self, idx):
        key = str(self.what.itemData(idx) or '')
        if key == 'blank':
            self.actions_widget.setVisible(False)
            self.spacer_widget.setVisible(True)
        else:
            self.actions_widget.setVisible(True)
            self.spacer_widget.setVisible(False)
            self.all_actions.setModel(self.models[key][0])
            self.current_actions.setModel(self.models[key][1])

    def add_action(self, *args):
        all_sm = self.all_actions.selectionModel()
        assert all_sm is not None
        self._add_action(all_sm.selectedIndexes())

    def add_single_action(self, index):
        self._add_action([index])

    def _add_action(self, indices):
        all_model = self.all_actions.model()
        assert isinstance(all_model, AllModel)
        names = all_model.names(indices)
        if names:
            cur_model = self.current_actions.model()
            assert isinstance(cur_model, CurrentModel)
            not_added = cur_model.add(names)
            ns = {y.name for y in not_added}
            added = set(names) - ns
            all_model.remove(indices, added)
            if not_added:
                warning_dialog(
                    self,
                    _('Cannot add'),
                    _('Cannot add the actions %s to this location') % ','.join([a.action_spec[0] for a in not_added]),
                    show=True,
                )
            if added:
                ca = self.current_actions
                ca_model = ca.model()
                assert isinstance(ca_model, CurrentModel)
                idx = ca_model.index(ca_model.rowCount() - 1, 0)
                ca.scrollTo(idx)
                self.changed_signal.emit()

    def remove_action(self, *args):
        cur_sm = self.current_actions.selectionModel()
        assert cur_sm is not None
        self._remove_action(cur_sm.selectedIndexes())

    def remove_single_action(self, index):
        self._remove_action([index])

    def _remove_action(self, indices):
        cur_model = self.current_actions.model()
        assert isinstance(cur_model, CurrentModel)
        names = cur_model.names(indices)
        if names:
            not_removed = cur_model.remove(indices)
            ns = {y.name for y in not_removed}
            removed = set(names) - ns
            all_model = self.all_actions.model()
            assert isinstance(all_model, AllModel)
            all_model.add(removed)
            if not_removed:
                warning_dialog(
                    self,
                    _('Cannot remove'),
                    _('Cannot remove the actions %s from this location') % ','.join([a.action_spec[0] for a in not_removed]),
                    show=True,
                )
            else:
                self.changed_signal.emit()

    def _move_action(self, delta: int) -> None:
        sm = self.current_actions.selectionModel()
        assert sm is not None
        x = sm.selectedIndexes()
        if x and len(x):
            i = sm.currentIndex().row()
            m = self.current_actions.model()
            assert isinstance(m, CurrentModel)
            idx_map = m.move_many(x, delta)
            newci = idx_map.get(i)
            sm.clear()
            for idx in idx_map.values():
                sm.select(idx, QItemSelectionModel.SelectionFlag.Select)
            if newci is not None:
                sm.setCurrentIndex(newci, QItemSelectionModel.SelectionFlag.SelectCurrent)
            if newci is not None:
                self.current_actions.scrollTo(newci, QAbstractItemView.ScrollHint.EnsureVisible)
            self.changed_signal.emit()

    def commit(self, *args):
        # Ensure preferences are showing in either the toolbar or
        # the menubar.
        pref_in_toolbar = self.models['toolbar'][1].has_action('Preferences')
        pref_in_menubar = self.models['menubar'][1].has_action('Preferences')
        lm_in_toolbar = self.models['toolbar-device'][1].has_action('Location Manager')
        lm_in_menubar = self.models['menubar-device'][1].has_action('Location Manager')
        if not pref_in_toolbar and not pref_in_menubar:
            error_dialog(
                self,
                _('Preferences missing'),
                _('The Preferences action must be in either the main toolbar or the menubar.'),
                show=True,
            )
            raise AbortCommit()
        if not lm_in_toolbar and not lm_in_menubar:
            error_dialog(
                self,
                _('Location manager missing'),
                _('The Location manager must be in either the main toolbar or the menubar when a device is connected.'),
                show=True,
            )
            raise AbortCommit()

        # Save data.
        for am, cm in self.models.values():
            cm.commit()
        return False

    def restore_defaults(self, *args):
        for am, cm in self.models.values():
            cm.restore_defaults()
            am.restore_defaults()
        self.changed_signal.emit()

    def refresh_gui(self, gui):
        gui.bars_manager.init_bars()
        gui.bars_manager.update_bars()


if __name__ == '__main__':
    from calibre.gui2 import Application

    app = Application([])
    test_widget('Interface', 'Toolbar')



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

from qt.core import QUrl

from calibre.gui2 import open_url
from calibre.gui2.actions import InterfaceAction
from calibre.utils.localization import _, localize_user_manual_link


class HelpAction(InterfaceAction):
    name = 'Help'
    action_spec = (
        _('Help'),
        'help.png',
        _('Browse the calibre User Manual'),
        _('F1'),
    )

    def genesis(self):
        self.qaction.triggered.connect(self.show_help)

    def show_help(self, *args):
        open_url(QUrl(localize_user_manual_link('https://manual.calibre-ebook.com')))



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

import os
import re
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Protocol

from qt.core import (
    QAction,
    QApplication,
    QClipboard,
    QColor,
    QComboBox,
    QCompleter,
    QContextMenuEvent,
    QCursor,
    QEvent,
    QFont,
    QGraphicsPixmapItem,
    QGraphicsScene,
    QGraphicsView,
    QIcon,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QPageSize,
    QPainter,
    QPalette,
    QPen,
    QPixmap,
    QPrinter,
    QRect,
    QSize,
    QSplitterHandle,
    QStringListModel,
    QSyntaxHighlighter,
    Qt,
    QTextCharFormat,
    QWidget,
    pyqtSignal,
)

from calibre import fit_image, force_unicode, strftime
from calibre.constants import ismacos, iswindows
from calibre.ebooks import BOOK_EXTENSIONS
from calibre.gui2 import clip_border_radius, error_dialog, gprefs, pixmap_to_data, qapplication_or_fail, warning_dialog
from calibre.gui2.dnd import DownloadDialog, dnd_get_files, dnd_get_image, dnd_get_local_image_and_pixmap, dnd_has_extension, dnd_has_image, image_extensions
from calibre.gui2.filename_pattern_ui import Ui_Form
from calibre.gui2.progress_indicator import ProgressIndicator as _ProgressIndicator
from calibre.startup import connect_lambda
from calibre.utils.config import XMLConfig, prefs
from calibre.utils.localization import _, localize_user_manual_link

history = XMLConfig('history')


class ProgressIndicator(QWidget):  # {{{
    def __init__(self, *args):
        QWidget.__init__(self, *args)
        self.setGeometry(0, 0, 300, 350)
        self.pi = _ProgressIndicator(self)
        self.status = QLabel(self)
        self.status.setWordWrap(True)
        self.status.setAlignment(Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop)
        self.setVisible(False)
        self.current_pos: tuple[int, int] | None = None

    def start(self, msg=''):
        view = self.parent()
        assert isinstance(view, QWidget)
        pwidth, pheight = view.size().width(), view.size().height()
        self.resize(pwidth, min(pheight, 250))
        if self.current_pos is None:
            self.move(0, int((pheight - self.size().height()) / 2))
        else:
            self.move(self.current_pos[0], self.current_pos[1])
        self.pi.resize(self.pi.sizeHint())
        self.pi.move(int((self.size().width() - self.pi.size().width()) / 2), 0)
        self.status.resize(self.size().width(), self.size().height() - self.pi.size().height() - 10)
        self.status.move(0, self.pi.size().height() + 10)
        self.status.setText('<h1>' + msg + '</h1>')
        self.setVisible(True)
        self.pi.startAnimation()

    def stop(self):
        self.pi.stopAnimation()
        self.setVisible(False)


# }}}


class FilenamePattern(QWidget, Ui_Form):  # {{{
    changed_signal = pyqtSignal()

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.setupUi(self)
        try:
            self.help_label.setText(self.help_label.text() % localize_user_manual_link('https://manual.calibre-ebook.com/regexp.html'))
        except TypeError:
            pass  # link already localized

        self.test_button.clicked.connect(self.do_test)
        re_line_edit = self.re.lineEdit()
        assert re_line_edit is not None
        re_line_edit.returnPressed[()].connect(self.do_test)
        self.filename.returnPressed[()].connect(self.do_test)
        connect_lambda(re_line_edit.textChanged, self, lambda self, x: self.changed_signal.emit())

    def initialize(self, defaults=False):
        # Get all items in the combobox. If we are resetting
        # to defaults we don't want to lose what the user
        # has added.
        re_line_edit = self.re.lineEdit()
        assert re_line_edit is not None
        val_hist = [str(re_line_edit.text())] + [str(self.re.itemText(i)) for i in range(self.re.count())]
        self.re.clear()

        if defaults:
            val = prefs.defaults['filename_pattern']
        else:
            val = prefs['filename_pattern']
        re_line_edit.setText(val)

        val_hist += gprefs.get(
            'filename_pattern_history',
            [
                '(?P<title>.+)',
                r'(?P<author>[^_-]+) -?\s*(?P<series>[^_0-9-]*)(?P<series_index>[0-9]*)\s*-\s*(?P<title>[^_].+) ?',
            ],
        )
        if val in val_hist:
            del val_hist[val_hist.index(val)]
        val_hist.insert(0, val)
        for v in val_hist:
            # Ensure we don't have duplicate items.
            if v and self.re.findText(v) == -1:
                self.re.addItem(v)
        self.re.setCurrentIndex(0)

    def do_test(self):
        from calibre.ebooks.metadata import authors_to_string
        from calibre.ebooks.metadata.meta import metadata_from_filename

        fname = str(self.filename.text())
        ext = os.path.splitext(fname)[1][1:].lower()
        if ext not in BOOK_EXTENSIONS:
            return warning_dialog(
                self,
                _('Test file name invalid'),
                _('The file name <b>%s</b> does not appear to end with a file extension. It must end with a file  extension like .epub or .mobi') % fname,
                show=True,
            )

        try:
            pat = self.pattern()
        except Exception as err:
            error_dialog(self, _('Invalid regular expression'), _('Invalid regular expression: %s') % err).exec()
            return
        mi = metadata_from_filename(fname, pat)
        if mi.title:
            self.title.setText(mi.title)
        else:
            self.title.setText(_('No match'))
        if mi.authors:
            self.authors.setText(authors_to_string(mi.authors))
        else:
            self.authors.setText(_('No match'))

        if mi.series:
            self.series.setText(mi.series)
        else:
            self.series.setText(_('No match'))

        if mi.series_index is not None:
            self.series_index.setText(str(mi.series_index))
        else:
            self.series_index.setText(_('No match'))

        if mi.publisher:
            self.publisher.setText(mi.publisher)
        else:
            self.publisher.setText(_('No match'))

        if mi.pubdate:
            self.pubdate.setText(strftime('%Y-%m-%d', mi.pubdate))
        else:
            self.pubdate.setText(_('No match'))

        self.isbn.setText(_('No match') if mi.isbn is None else str(mi.isbn))
        self.comments.setText(mi.comments or _('No match'))

    def pattern(self):
        re_line_edit = self.re.lineEdit()
        assert re_line_edit is not None
        pat = str(re_line_edit.text())
        return re.compile(pat)

    def commit(self):
        pat = self.pattern().pattern
        prefs['filename_pattern'] = pat

        history = []
        re_line_edit = self.re.lineEdit()
        assert re_line_edit is not None
        history_pats = [str(re_line_edit.text())] + [str(self.re.itemText(i)) for i in range(self.re.count())]
        for p in history_pats[:24]:
            # Ensure we don't have duplicate items.
            if p and p not in history:
                history.append(p)
        gprefs['filename_pattern_history'] = history

        return pat


# }}}


class FormatList(QListWidget):  # {{{
    DROPABBLE_EXTENSIONS = BOOK_EXTENSIONS
    formats_dropped = pyqtSignal(object, object)
    delete_format = pyqtSignal()

    def dragEnterEvent(self, e):
        md = e.mimeData()
        if dnd_has_extension(md, self.DROPABBLE_EXTENSIONS, allow_all_extensions=True):
            e.acceptProposedAction()

    def dropEvent(self, event):
        event.setDropAction(Qt.DropAction.CopyAction)
        md = event.mimeData()
        # Now look for ebook files
        urls, filenames = dnd_get_files(md, self.DROPABBLE_EXTENSIONS, allow_all_extensions=True)
        if not urls:
            # Nothing found
            return

        if not filenames:
            # Local files
            self.formats_dropped.emit(event, urls)
        else:
            # Remote files, use the first file
            d = DownloadDialog(urls[0], filenames[0], self)
            d.start_download()
            if d.err is None:
                self.formats_dropped.emit(event, [d.fpath])

    def dragMoveEvent(self, e):
        e.acceptProposedAction()

    def keyPressEvent(self, e):
        if e.key() == Qt.Key.Key_Delete:
            self.delete_format.emit()
        else:
            return QListWidget.keyPressEvent(self, e)


# }}}


class ImageDropWidgetProtocol(Protocol):  # {{{
    def setAcceptDrops(self, on: bool) -> None: ...
    def pixmap(self) -> QPixmap: ...
    def get_pixmap(self) -> QPixmap: ...
    def setPixmap(self, pixmap: QPixmap) -> None: ...
    def set_pixmap(self, pixmap: QPixmap) -> None: ...
    def handle_image_drop(self, pmap: QPixmap, data: bytes | None = None) -> None: ...

    cover_changed: Any


class ImageDropMixin:
    """
    Adds support for dropping images onto widgets and a context menu for
    copy/pasting images.
    """

    DROPABBLE_EXTENSIONS = None

    if TYPE_CHECKING:

        def pixmap(self) -> QPixmap: ...
        def setPixmap(self, pixmap: QPixmap) -> None: ...

    def __init__(self: ImageDropWidgetProtocol):
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        md = event.mimeData()
        exts = self.DROPABBLE_EXTENSIONS or image_extensions()
        if dnd_has_extension(md, exts) or dnd_has_image(md):
            event.acceptProposedAction()

    def dropEvent(self: ImageDropWidgetProtocol, event):
        event.setDropAction(Qt.DropAction.CopyAction)
        md = event.mimeData()
        pmap, data = dnd_get_local_image_and_pixmap(md)
        if pmap is not None:
            self.handle_image_drop(pmap, data)
            return

        x, y = dnd_get_image(md)
        if x is not None:
            # We have an image, set cover
            event.accept()
            if y is None:
                # Local image
                self.handle_image_drop(x)
            else:
                # Remote files, use the first file
                d = DownloadDialog(x, y, self)
                d.start_download()
                if d.err is None:
                    pmap = QPixmap()
                    with open(d.fpath, 'rb') as f:
                        data = f.read()
                    pmap.loadFromData(data)
                    if not pmap.isNull():
                        self.handle_image_drop(pmap, data=data)

    def handle_image_drop(self: ImageDropWidgetProtocol, pmap, data=None):
        self.set_pixmap(pmap)
        self.cover_changed.emit(data or pixmap_to_data(pmap, format='PNG'))

    def dragMoveEvent(self, event):
        event.acceptProposedAction()

    def get_pixmap(self) -> QPixmap:
        return self.pixmap()

    def set_pixmap(self, pixmap: QPixmap) -> None:
        self.setPixmap(pixmap)

    def build_context_menu(self):
        cm = QMenu(self)  # type: ignore
        paste = cm.addAction(QIcon.ic('edit-paste.png'), _('Paste cover'))
        assert paste is not None
        copy = cm.addAction(QIcon.ic('edit-copy.png'), _('Copy cover'))
        assert copy is not None
        _cb = qapplication_or_fail().clipboard()
        assert _cb is not None
        _mime = _cb.mimeData()
        assert _mime is not None
        if not _mime.hasImage():
            paste.setEnabled(False)
        copy.triggered.connect(self.copy_to_clipboard)
        paste.triggered.connect(self.paste_from_clipboard)
        return cm

    def contextMenuEvent(self, a0):
        self.build_context_menu().exec(a0.globalPos())

    def copy_to_clipboard(self) -> None:
        _cb = qapplication_or_fail().clipboard()
        assert _cb is not None
        _cb.setPixmap(self.get_pixmap())

    def paste_from_clipboard(self: ImageDropWidgetProtocol):
        cb = qapplication_or_fail().clipboard()
        assert cb is not None
        pmap = cb.pixmap()
        if pmap.isNull() and cb.supportsSelection():
            pmap = cb.pixmap(QClipboard.Mode.Selection)
        if ismacos:  # Without this there is a crash when Qt tries to save this pixmap as JPEG data
            pmap = pmap.copy()
        if not pmap.isNull():
            self.set_pixmap(pmap)
            self.cover_changed.emit(pixmap_to_data(pmap, format='PNG'))
            return True
        return False


# }}}

# ImageView {{{


def draw_size(p, rect, w, h):
    rect = rect.adjusted(0, 0, 0, -4)
    f = p.font()
    f.setBold(True)
    p.setFont(f)
    sz = f'\xa0{w} x {h}\xa0'
    flags = Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight | Qt.TextFlag.TextSingleLine
    szrect = p.boundingRect(rect, flags, sz)
    p.fillRect(szrect.adjusted(0, 0, 0, 4), QColor(0, 0, 0, 200))
    p.setPen(QPen(QColor(255, 255, 255)))
    p.drawText(rect, flags, sz)


class ImageView(QWidget, ImageDropMixin):
    BORDER_WIDTH = 1
    cover_changed = pyqtSignal(object)
    draw_empty_border = False

    def __init__(self, parent=None, show_size_pref_name=None, default_show_size=False):
        QWidget.__init__(self, parent)
        self.show_size_pref_name = ('show_size_on_cover_' + show_size_pref_name) if show_size_pref_name else None
        self._pixmap = QPixmap()
        self.setMinimumSize(QSize(150, 200))
        ImageDropMixin.__init__(self)
        self.draw_border = True
        self.show_size = False
        if self.show_size_pref_name:
            self.show_size = gprefs.get(self.show_size_pref_name, default_show_size)

    def dragEnterEvent(self, a0):
        ImageDropMixin.dragEnterEvent(self, a0)

    def dragMoveEvent(self, a0):
        ImageDropMixin.dragMoveEvent(self, a0)

    def dropEvent(self, a0):
        ImageDropMixin.dropEvent(self, a0)

    def setPixmap(self, pixmap):
        if not isinstance(pixmap, QPixmap):
            raise TypeError('Must use a QPixmap')
        self._pixmap = pixmap
        self.updateGeometry()
        self.update()

    def build_context_menu(self):
        m = ImageDropMixin.build_context_menu(self)
        if self.show_size_pref_name:
            text = _('Hide size in corner') if self.show_size else _('Show size in corner')
            m.addAction(text, self.toggle_show_size)
        return m

    def toggle_show_size(self):
        self.show_size ^= True
        if self.show_size_pref_name:
            gprefs[self.show_size_pref_name] = self.show_size
        self.update()

    def pixmap(self):
        return self._pixmap

    def sizeHint(self):
        if self._pixmap.isNull():
            return self.minimumSize()
        return self._pixmap.size()

    def paintEvent(self, a0):
        QWidget.paintEvent(self, a0)
        pmap = self._pixmap
        p = QPainter(self)
        p.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)
        if pmap.isNull():
            if self.draw_empty_border:
                pen = QPen()
                pen.setWidth(self.BORDER_WIDTH)
                p.setPen(pen)
                p.drawRect(self.rect())
                p.end()
            return
        w, h = pmap.width(), pmap.height()
        ow, oh = w, h
        cw, ch = self.rect().width(), self.rect().height()
        scaled, nw, nh = fit_image(w, h, cw, ch)
        if scaled:
            pmap = pmap.scaled(
                int(nw * pmap.devicePixelRatio()),
                int(nh * pmap.devicePixelRatio()),
                Qt.AspectRatioMode.IgnoreAspectRatio,
                Qt.TransformationMode.SmoothTransformation,
            )
        w, h = int(pmap.width() / pmap.devicePixelRatio()), int(pmap.height() / pmap.devicePixelRatio())
        x = int(abs(cw - w) / 2)
        y = int(abs(ch - h) / 2)
        target = QRect(x, y, w, h)
        with clip_border_radius(p, target):
            p.drawPixmap(target, pmap)
        if self.draw_border:
            pen = QPen()
            pen.setWidth(self.BORDER_WIDTH)
            p.setPen(pen)
            p.drawRect(target)
        if self.show_size:
            draw_size(p, target, ow, oh)


# }}}

# CoverView {{{


class RoundedPixmap(QGraphicsPixmapItem):
    def paint(self, painter, option, widget):
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        target = self.boundingRect().toAlignedRect()
        with clip_border_radius(painter, target):
            painter.drawPixmap(target, self.pixmap())


class CoverView(QGraphicsView, ImageDropMixin):
    cover_changed = pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        self.show_size = kwargs.pop('show_size', False)
        QGraphicsView.__init__(self, *args, **kwargs)
        ImageDropMixin.__init__(self)
        self.pixmap_size = 0, 0
        self._pixmap_scene: QGraphicsScene | None = None
        if self.show_size:
            self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.FullViewportUpdate)
        self.set_background()

    def contextMenuEvent(self, event):
        return ImageDropMixin.contextMenuEvent(self, event)

    def get_pixmap(self) -> QPixmap:
        if self._pixmap_scene is not None:
            for item in self._pixmap_scene.items():
                if isinstance(item, QGraphicsPixmapItem):
                    return item.pixmap()
        return QPixmap()

    def pixmap(self) -> QPixmap:
        return self.get_pixmap()

    def set_pixmap(self, pixmap: QPixmap) -> None:
        self._pixmap_scene = QGraphicsScene()
        self._pixmap_scene.addItem(RoundedPixmap(pixmap))
        self.setScene(self._pixmap_scene)

    def setPixmap(self, pixmap: QPixmap) -> None:
        self.set_pixmap(pixmap)

    def set_background(self, brush=None):
        self.setBackgroundBrush(brush or self.palette().color(QPalette.ColorRole.Window))

    def paintEvent(self, event):
        QGraphicsView.paintEvent(self, event)
        if self.show_size:
            v = self.viewport()
            assert v is not None
            p = QPainter(v)
            draw_size(p, v.rect(), *self.pixmap_size)


# }}}

# BasicList {{{


class BasicListItem(QListWidgetItem):
    def __init__(self, text, user_data=None):
        QListWidgetItem.__init__(self, text)
        self.user_data = user_data

    def __eq__(self, other):
        if hasattr(other, 'text'):
            return self.text() == other.text()
        return False


class BasicList(QListWidget):
    def add_item(self, text, user_data=None, replace=False):
        item = BasicListItem(text, user_data)

        for oitem in self.items():
            if oitem == item:
                if replace:
                    self.takeItem(self.row(oitem))
                else:
                    raise ValueError('Item already in list')

        self.addItem(item)

    def remove_selected_items(self, *args):
        for item in self.selectedItems():
            self.takeItem(self.row(item))

    def items(self, data=None):
        for i in range(self.count()):
            yield self.item(i)


# }}}


class LineEditECMProtocol(Protocol):
    def createStandardContextMenu(self) -> QMenu | None: ...
    def create_change_case_menu(self, menu: QMenu) -> QMenu: ...
    def add_items_to_context_menu(self, menu: QMenu) -> QMenu: ...
    def hasSelectedText(self) -> bool: ...
    def selectedText(self) -> str: ...
    def text(self) -> str: ...
    def insert(self, text: str) -> None: ...
    def setText(self, text: str) -> None: ...
    def modify_case_operation(self, func: Callable[[str], str]) -> None: ...


class LineEditECM:  # {{{
    """
    Extend the context menu of a QLineEdit to include more actions.
    """

    add_items_to_context_menu_callback: Callable[[QMenu], None] | None = None

    if TYPE_CHECKING:

        def hasSelectedText(self) -> bool: ...
        def selectedText(self) -> str: ...
        def text(self) -> str: ...
        def insert(self, a0: str | None) -> None: ...
        def setText(self, a0: str | None) -> None: ...
        def createStandardContextMenu(self) -> QMenu | None: ...

    def create_change_case_menu(self, menu: QMenu) -> QMenu:
        case_menu = QMenu(_('Change case'), menu)
        action_upper_case = case_menu.addAction(_('Upper case'))
        action_lower_case = case_menu.addAction(_('Lower case'))
        action_swap_case = case_menu.addAction(_('Swap case'))
        action_title_case = case_menu.addAction(_('Title case'))
        action_capitalize = case_menu.addAction(_('Capitalize'))
        assert action_upper_case is not None
        assert action_lower_case is not None
        assert action_swap_case is not None
        assert action_title_case is not None
        assert action_capitalize is not None
        action_upper_case.triggered.connect(self.upper_case)
        action_lower_case.triggered.connect(self.lower_case)
        action_swap_case.triggered.connect(self.swap_case)
        action_title_case.triggered.connect(self.title_case)
        action_capitalize.triggered.connect(self.capitalize)
        menu.addMenu(case_menu)
        return case_menu

    def add_items_to_context_menu(self, menu: QMenu) -> QMenu:
        if self.add_items_to_context_menu_callback is not None:
            self.add_items_to_context_menu_callback(menu)
        return menu

    def contextMenuEvent(self, a0: QContextMenuEvent | None) -> None:
        menu = self.createStandardContextMenu()
        assert isinstance(menu, QMenu)
        menu.addSeparator()
        self.create_change_case_menu(menu)
        menu = self.add_items_to_context_menu(menu)
        menu.exec(a0.globalPos())  # type: ignore

    def modify_case_operation(self, func: Callable[[str], str]) -> None:
        has_selection = self.hasSelectedText()
        text = self.selectedText() if has_selection else self.text()
        ntext = func(text)
        if ntext != text:
            self.insert(ntext) if has_selection else self.setText(ntext)

    def upper_case(self: LineEditECMProtocol):
        from calibre.utils.icu import upper

        self.modify_case_operation(upper)

    def lower_case(self: LineEditECMProtocol):
        from calibre.utils.icu import lower

        self.modify_case_operation(lower)

    def swap_case(self: LineEditECMProtocol):
        from calibre.utils.icu import swapcase

        self.modify_case_operation(swapcase)

    def title_case(self: LineEditECMProtocol):
        from calibre.utils.titlecase import titlecase

        self.modify_case_operation(titlecase)

    def capitalize(self: LineEditECMProtocol):
        from calibre.utils.icu import capitalize

        self.modify_case_operation(capitalize)


# }}}


class EnLineEdit(LineEditECM, QLineEdit):  # {{{
    """
    Enhanced QLineEdit.

    Includes an extended content menu.
    """

    def event(self, a0):
        # See https://bugreports.qt.io/browse/QTBUG-46911
        if a0.type() == QEvent.Type.ShortcutOverride and (
            hasattr(a0, 'key')
            and a0.key() in (Qt.Key.Key_Left, Qt.Key.Key_Right)
            and (a0.modifiers() & ~Qt.KeyboardModifier.KeypadModifier) == Qt.KeyboardModifier.ControlModifier
        ):
            a0.accept()
        return QLineEdit.event(self, a0)


# }}}

# LineEditIndicators {{{


class LineEditIndicatorsProtocol(Protocol):
    status_actions: tuple[QAction, QAction]

    def addAction(self, icon: QIcon, position: QLineEdit.ActionPosition) -> QAction | None: ...
    def setStyleSheet(self, styleSheet: str) -> None: ...


def setup_status_actions(self: LineEditIndicatorsProtocol):
    ok = self.addAction(QIcon.ic('ok.png'), QLineEdit.ActionPosition.TrailingPosition)
    err = self.addAction(QIcon.ic('dialog_error.png'), QLineEdit.ActionPosition.TrailingPosition)
    assert ok is not None and err is not None
    self.status_actions = ok, err
    self.status_actions[0].setVisible(False)
    self.status_actions[1].setVisible(False)


def stylesheet_for_lineedit(ok, selector='QLineEdit') -> str:
    if ok is None:
        return ''
    col = '#50c878' if ok else '#FF2400'
    return f'{selector} {{ border: 2px solid {col}; border-radius: 3px }}'


def update_status_actions(self: LineEditIndicatorsProtocol, ok, tooltip: str = ''):
    self.status_actions[0].setVisible(bool(ok))
    self.status_actions[1].setVisible(not ok)
    if ok:
        self.status_actions[0].setToolTip(tooltip)
    elif ok is None:
        self.status_actions[1].setVisible(False)
    else:
        self.status_actions[1].setToolTip(tooltip)
    self.setStyleSheet(stylesheet_for_lineedit(ok))


class LineEditIndicators:
    status_actions: tuple[QAction, QAction]

    def setup_status_actions(self: LineEditIndicatorsProtocol):
        setup_status_actions(self)

    def update_status_actions(self: LineEditIndicatorsProtocol, ok, tooltip=''):
        update_status_actions(self, ok, tooltip)


# }}}


class ItemsCompleter(QCompleter):  # {{{
    """
    A completer object that completes a list of tags. It is used in conjunction
    with a CompleterLineEdit.
    """

    def __init__(self, parent, all_items):
        QCompleter.__init__(self, all_items, parent)
        self.all_items = set(all_items)

    def update(self, text_items, completion_prefix):
        items = list(self.all_items.difference(text_items))
        model = QStringListModel(items, self)
        self.setModel(model)

        self.setCompletionPrefix(completion_prefix)
        if completion_prefix.strip():
            self.complete()

    def update_items_cache(self, items):
        self.all_items = set(items)
        model = QStringListModel(items, self)
        self.setModel(model)


# }}}


class CompleteLineEdit(EnLineEdit):  # {{{
    """
    A QLineEdit that can complete parts of text separated by separator.
    """

    def __init__(self, parent=None, complete_items=[], sep=',', space_before_sep=False):
        EnLineEdit.__init__(self, parent)

        self.separator = sep
        self.space_before_sep = space_before_sep

        self.textChanged.connect(self.text_changed)

        self.items_completer = ItemsCompleter(self, complete_items)
        self.items_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)

        self.items_completer.activated[str].connect(self.complete_text)

        self.items_completer.setWidget(self)

    def update_items_cache(self, complete_items):
        self.items_completer.update_items_cache(complete_items)

    def set_separator(self, sep):
        self.separator = sep

    def set_space_before_sep(self, space_before):
        self.space_before_sep = space_before

    def text_changed(self, text):
        all_text = str(text)
        text = all_text[: self.cursorPosition()]
        prefix = text.split(self.separator)[-1].strip()

        text_items = []
        for t in all_text.split(self.separator):
            t1 = str(t).strip()
            if t1:
                text_items.append(t)
        text_items = list(set(text_items))
        self.items_completer.update(text_items, prefix)

    def complete_text(self, text):
        cursor_pos = self.cursorPosition()
        before_text = str(self.text())[:cursor_pos]
        after_text = str(self.text())[cursor_pos:]
        prefix_len = len(before_text.split(self.separator)[-1].lstrip())
        if self.space_before_sep:
            complete_text_pat = '%s%s %s %s'
            len_extra = 3
        else:
            complete_text_pat = '%s%s%s %s'
            len_extra = 2
        self.setText(complete_text_pat % (before_text[: cursor_pos - prefix_len], text, self.separator, after_text))
        self.setCursorPosition(cursor_pos - prefix_len + len(text) + len_extra)


# }}}


class EnComboBox(QComboBox):  # {{{
    """
    Enhanced QComboBox.

    Includes an extended context menu.
    """

    def __init__(self, *args):
        QComboBox.__init__(self, *args)
        self.setLineEdit(EnLineEdit(self))
        _completer = self.completer()
        assert _completer is not None
        _completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.setMinimumContentsLength(20)

    def text(self):
        return str(self.currentText())

    def setText(self, text):
        idx = self.findText(text, Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchCaseSensitive)
        if idx == -1:
            self.insertItem(0, text)
            idx = 0
        self.setCurrentIndex(idx)


# }}}


class CompleteComboBox(EnComboBox):  # {{{
    def __init__(self, *args):
        EnComboBox.__init__(self, *args)
        self.setLineEdit(CompleteLineEdit(self))

    def update_items_cache(self, complete_items):
        le = self.lineEdit()
        assert le is not None
        assert isinstance(le, CompleteLineEdit)
        le.update_items_cache(complete_items)

    def set_separator(self, sep):
        le = self.lineEdit()
        assert le is not None
        assert isinstance(le, CompleteLineEdit)
        le.set_separator(sep)

    def set_space_before_sep(self, space_before):
        le = self.lineEdit()
        assert le is not None
        assert isinstance(le, CompleteLineEdit)
        le.set_space_before_sep(space_before)


# }}}


class HistoryLineEdit(QComboBox):  # {{{
    lost_focus = pyqtSignal()

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.setEditable(True)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.setMaxCount(10)
        _line_edit = self.lineEdit()
        assert _line_edit is not None
        self.setClearButtonEnabled = _line_edit.setClearButtonEnabled
        self.textChanged = self.editTextChanged

    def setPlaceholderText(self, placeholderText):
        line_edit = self.lineEdit()
        assert line_edit is not None
        return line_edit.setPlaceholderText(placeholderText)

    def contextMenuEvent(self, e):
        _line_edit = self.lineEdit()
        assert _line_edit is not None
        menu = _line_edit.createStandardContextMenu()
        assert menu is not None
        menu.addSeparator()
        menu.addAction(_('Clear history'), self.clear_history_default_impl)
        menu.exec(e.globalPos())

    def clear_history_default_impl(self):
        self.clear()
        history.set(self.store_name, [])

    @property
    def store_name(self):
        return 'lineedit_history_' + self._name

    def initialize(self, name):
        self._name = name
        self.addItems(history.get(self.store_name, []))
        self.setEditText('')
        _line_edit = self.lineEdit()
        assert _line_edit is not None
        _line_edit.editingFinished.connect(self.save_history)

    def save_history(self):
        items = []
        ct = str(self.currentText())
        if ct:
            items.append(ct)
        for i in range(self.count()):
            item = str(self.itemText(i))
            if item not in items:
                items.append(item)
        self.blockSignals(True)
        self.clear()
        self.addItems(items)
        self.setEditText(ct)
        self.blockSignals(False)
        try:
            history.set(self.store_name, items)
        except ValueError:
            from calibre.utils.cleantext import clean_ascii_chars

            items = [clean_ascii_chars(force_unicode(x)) for x in items]
            try:
                history.set(self.store_name, items)
            except ValueError:
                pass

    def setText(self, t):
        self.setEditText(t)
        line_edit = self.lineEdit()
        assert line_edit is not None
        line_edit.setCursorPosition(0)

    def text(self):
        return self.currentText()

    def focusOutEvent(self, e):
        QComboBox.focusOutEvent(self, e)
        view = self.view()
        assert view is not None
        if not (self.hasFocus() or view.hasFocus()):
            self.lost_focus.emit()


# }}}


class ComboBoxWithHelp(QComboBox):  # {{{
    """
    A combobox where item 0 is help text. CurrentText will return '' for item 0.
    Be sure to always fetch the text with currentText. Don't use the signals
    that pass a string, because they will not correct the text.
    """

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.currentIndexChanged.connect(self.index_changed)
        self.help_text = ''
        self.state_set = False

    def initialize(self, help_text=_('Search')):
        self.help_text = help_text
        self.set_state()

    def set_state(self):
        if not self.state_set:
            if self.currentIndex() == 0:
                self.setItemText(0, self.help_text)
                self.setStyleSheet('QComboBox { color: gray }')
            else:
                self.setItemText(0, '')
                self.setStyleSheet('QComboBox { color: black }')

    def index_changed(self, index):
        self.state_set = False
        self.set_state()

    def currentText(self):
        if self.currentIndex() == 0:
            return ''
        return QComboBox.currentText(self)

    def itemText(self, index):
        if index == 0:
            return ''
        return QComboBox.itemText(self, index)

    def showPopup(self):
        self.setItemText(0, '')
        QComboBox.showPopup(self)

    def hidePopup(self):
        QComboBox.hidePopup(self)
        self.set_state()


# }}}


class EncodingComboBox(QComboBox):  # {{{
    """
    A combobox that holds text encodings support
    by Python. This is only populated with the most
    common and standard encodings. There is no good
    way to programmatically list all supported encodings
    using encodings.aliases.aliases.keys(). It
    will not work.
    """

    ENCODINGS = [
        '',
        'cp1252',
        'latin1',
        'utf-8',
        '',
        'ascii',
        'big5',
        'cp1250',
        'cp1251',
        'cp1253',
        'cp1254',
        'cp1255',
        'cp1256',
        'euc_jp',
        'euc_kr',
        'gb2312',
        'gb18030',
        'hz',
        'iso2022_jp',
        'iso2022_kr',
        'iso8859_5',
        'shift_jis',
    ]

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.setEditable(True)
        self.setLineEdit(EnLineEdit(self))

        for item in self.ENCODINGS:
            self.addItem(item)


# }}}


class PythonHighlighter(QSyntaxHighlighter):  # {{{
    Rules = ()
    Formats = {}

    KEYWORDS = [
        'and',
        'as',
        'assert',
        'break',
        'class',
        'continue',
        'def',
        'del',
        'elif',
        'else',
        'except',
        'exec',
        'finally',
        'for',
        'from',
        'global',
        'if',
        'import',
        'in',
        'is',
        'lambda',
        'not',
        'or',
        'pass',
        'print',
        'raise',
        'return',
        'try',
        'while',
        'with',
        'yield',
    ]

    BUILTINS = [
        'abs',
        'all',
        'any',
        'basestring',
        'bool',
        'callable',
        'chr',
        'classmethod',
        'cmp',
        'compile',
        'complex',
        'delattr',
        'dict',
        'dir',
        'divmod',
        'enumerate',
        'eval',
        'execfile',
        'exit',
        'file',
        'filter',
        'float',
        'frozenset',
        'getattr',
        'globals',
        'hasattr',
        'hex',
        'id',
        'int',
        'isinstance',
        'issubclass',
        'iter',
        'len',
        'list',
        'locals',
        'long',
        'map',
        'max',
        'min',
        'object',
        'oct',
        'open',
        'ord',
        'pow',
        'property',
        'range',
        'reduce',
        'repr',
        'reversed',
        'round',
        'set',
        'setattr',
        'slice',
        'sorted',
        'staticmethod',
        'str',
        'sum',
        'super',
        'tuple',
        'type',
        'unichr',
        'unicode',
        'vars',
        'xrange',
        'zip',
    ]

    CONSTANTS = ['False', 'True', 'None', 'NotImplemented', 'Ellipsis']

    def __init__(self, parent=None):
        super().__init__(parent)
        if not self.Rules:
            self.initialize_class_members()

    @classmethod
    def initialize_class_members(cls):
        cls.initializeFormats()
        r = []

        def a(a, b):
            r.append((a, b))

        a(re.compile('|'.join([rf'\b{keyword}\b' for keyword in cls.KEYWORDS])), 'keyword')
        a(re.compile('|'.join([rf'\b{builtin}\b' for builtin in cls.BUILTINS])), 'builtin')
        a(re.compile('|'.join([rf'\b{constant}\b' for constant in cls.CONSTANTS])), 'constant')
        a(
            re.compile(
                r'\b[+-]?[0-9]+[lL]?\b'
                r'|\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b'
                r'|\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b'
            ),
            'number',
        )
        a(re.compile(r'\bPyQt6\b|\bQt?[A-Z][a-z]\w+\b'), 'pyqt')
        a(re.compile(r'\b@\w+\b'), 'decorator')
        stringRe = re.compile(r'''(?:'[^']*?'|"[^"]*?")''')
        a(stringRe, 'string')
        cls.stringRe = re.compile(r"""(:?"["]".*?"["]"|'''.*?''')""")
        a(cls.stringRe, 'string')
        cls.tripleSingleRe = re.compile(r"""'''(?!")""")
        cls.tripleDoubleRe = re.compile(r'''"""(?!')''')
        cls.Rules = tuple(r)

    @classmethod
    def initializeFormats(cls):
        baseFormat = QTextCharFormat()
        baseFormat.setFontFamilies(['monospace'])
        p = qapplication_or_fail().palette()
        is_dark = qapplication_or_fail().is_dark_theme
        for name, color, bold, italic in (
            ('normal', None, False, False),
            ('keyword', p.color(QPalette.ColorRole.Link).name(), True, False),
            ('builtin', p.color(QPalette.ColorRole.Link).name(), False, False),
            ('constant', p.color(QPalette.ColorRole.Link).name(), False, False),
            ('decorator', '#0000E0', False, False),
            ('comment', '#00c700' if is_dark else '#007F00', False, True),
            ('string', '#b6b600' if is_dark else '#808000', False, False),
            ('number', '#d96d00' if is_dark else '#924900', False, False),
            ('error', '#FF0000', False, False),
            ('pyqt', '#50621A', False, False),
        ):
            fmt = QTextCharFormat(baseFormat)
            if color is not None:
                fmt.setForeground(QColor(color))
            if bold:
                fmt.setFontWeight(QFont.Weight.Bold)
            if italic:
                fmt.setFontItalic(italic)
            cls.Formats[name] = fmt

    def highlightBlock(self, text):
        NORMAL, TRIPLESINGLE, TRIPLEDOUBLE, ERROR = range(4)

        textLength = len(text)
        prevState = self.previousBlockState()

        self.setFormat(0, textLength, self.Formats['normal'])

        if text.startswith(('Traceback', 'Error: ')):
            self.setCurrentBlockState(ERROR)
            self.setFormat(0, textLength, self.Formats['error'])
            return
        if prevState == ERROR and not text.startswith(('>>>', '#')):
            self.setCurrentBlockState(ERROR)
            self.setFormat(0, textLength, self.Formats['error'])
            return

        for regex, fmt in PythonHighlighter.Rules:
            for m in regex.finditer(text):
                self.setFormat(m.start(), m.end() - m.start(), self.Formats[fmt])

        # Slow but good quality highlighting for comments. For more
        # speed, comment this out and add the following to __init__:
        # PythonHighlighter.Rules.append((re.compile(r"#.*"), "comment"))
        if not text:
            pass
        elif text[0] == '#':
            self.setFormat(0, len(text), self.Formats['comment'])
        else:
            stack = []
            for i, c in enumerate(text):
                if c in ('"', "'"):
                    if stack and stack[-1] == c:
                        stack.pop()
                    else:
                        stack.append(c)
                elif c == '#' and len(stack) == 0:
                    self.setFormat(i, len(text), self.Formats['comment'])
                    break

        self.setCurrentBlockState(NORMAL)

        if self.stringRe.search(text) is not None:
            return
        # This is fooled by triple quotes inside single quoted strings
        for m, state in (
            (self.tripleSingleRe.search(text), TRIPLESINGLE),
            (self.tripleDoubleRe.search(text), TRIPLEDOUBLE),
        ):
            i = -1 if m is None else m.start()
            if self.previousBlockState() == state:
                if i == -1:
                    i = len(text)
                    self.setCurrentBlockState(state)
                self.setFormat(0, i + 3, self.Formats['string'])
            elif i > -1:
                self.setCurrentBlockState(state)
                self.setFormat(i, len(text), self.Formats['string'])

    def rehighlight(self):
        QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))
        super().rehighlight()
        QApplication.restoreOverrideCursor()


# }}}

# Splitter {{{


class SplitterHandle(QSplitterHandle):
    double_clicked = pyqtSignal(object)

    def __init__(self, orientation, splitter):
        super().__init__(orientation, splitter)
        splitter.splitterMoved.connect(self.splitter_moved, type=Qt.ConnectionType.QueuedConnection)
        self.double_clicked.connect(splitter.double_clicked, type=Qt.ConnectionType.QueuedConnection)
        self.highlight = False
        self.setToolTip(_('Drag to resize') + ' ' + splitter.label)

    def splitter_moved(self, *args):
        oh = self.highlight
        splitter = self.splitter()
        assert splitter is not None
        self.highlight = 0 in splitter.sizes()
        if oh != self.highlight:
            self.update()

    def mouseDoubleClickEvent(self, a0):
        self.double_clicked.emit(self)


class PaperSizes(QComboBox):  # {{{
    system_default_paper_size = None

    def initialize(self, choices=None):
        from calibre.utils.icu import numeric_sort_key

        if self.system_default_paper_size is None:
            PaperSizes.system_default_paper_size = 'a4'
            if iswindows or ismacos:
                # On Linux, this can cause Qt to load the system cups plugin
                # which can crash: https://bugs.launchpad.net/calibre/+bug/1861741
                ps_id = QPrinter().pageLayout().pageSize().id()  # type: ignore
                PaperSizes.system_default_paper_size = 'letter' if ps_id == QPageSize.PageSizeId.Letter else 'a4'
        if not choices:
            from calibre.ebooks.conversion.plugins.pdf_output import PAPER_SIZES

            choices = PAPER_SIZES
        for a in sorted(choices, key=numeric_sort_key):
            s = getattr(QPageSize.PageSizeId, a.capitalize())
            sz = QPageSize.definitionSize(s)
            unit = {QPageSize.Unit.Millimeter: 'mm', QPageSize.Unit.Inch: 'inch'}[QPageSize.definitionUnits(s)]
            name = f'{QPageSize.name(s)} ({sz.width():g} x {sz.height():g} {unit})'
            self.addItem(name, a)

    @property
    def get_value_for_config(self):
        return self.currentData()

    @get_value_for_config.setter
    def set_value_for_config(self, val):
        idx = self.findData(val or PaperSizes.system_default_paper_size)
        if idx == -1:
            idx = self.findData('a4')
        self.setCurrentIndex(idx)


# }}}


class BusyCursor:  # {{{
    def __enter__(self):
        QApplication.setOverrideCursor(QCursor(Qt.CursorShape.WaitCursor))

    def __exit__(self, *args):
        QApplication.restoreOverrideCursor()


# }}}

if __name__ == '__main__':
    from qt.core import QTextEdit

    app = QApplication([])
    w = QTextEdit()
    s = PythonHighlighter(w)
    # w.setSyntaxHighlighter(s)
    w.setText(open(__file__, 'rb').read().decode('utf-8'))
    w.show()
    app.exec()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid at kovidgoyal.net>

import datetime
import hashlib
from collections import OrderedDict
from collections.abc import Callable
from functools import lru_cache, partial
from typing import Any, NamedTuple, cast
from urllib.parse import urlencode

from html5_parser import parse
from lxml import etree
from lxml.builder import ElementMaker

from calibre import force_unicode, guess_type
from calibre import prepare_string_for_xml as xml
from calibre.constants import __appname__
from calibre.db.cache import Cache
from calibre.db.categories import Tag
from calibre.db.view import sanitize_sort_field_name
from calibre.ebooks.metadata import authors_to_string, fmt_sidx, rating_to_stars
from calibre.library.comments import comments_to_html
from calibre.srv.errors import HTTPInternalServerError, HTTPNotFound
from calibre.srv.handler import Context, UrlForCallable
from calibre.srv.http_request import parse_uri
from calibre.srv.http_response import RequestData
from calibre.srv.opts import Options
from calibre.srv.routes import endpoint
from calibre.srv.utils import MultiDict, Offsets, get_library_data, http_date
from calibre.utils.config import prefs
from calibre.utils.date import as_utc, is_date_undefined, timestampfromdt
from calibre.utils.icu import sort_key
from calibre.utils.localization import _, ngettext
from calibre.utils.search_query_parser import ParseException
from calibre.utils.xml_parse import safe_xml_fromstring
from polyglot.binary import as_hex_unicode, from_hex_unicode
from polyglot.builtins import as_bytes
from polyglot.urllib import unquote_plus


class Group(NamedTuple):
    text: str
    count: int


def atom(ctx: Context, rd: RequestData, endpoint: Callable, output: bytes | str | etree.Element) -> bytes:
    rd.outheaders.set('Content-Type', 'application/atom+xml; charset=UTF-8', replace_all=True)
    rd.outheaders.set('Calibre-Instance-Id', force_unicode(prefs['installation_uuid'], 'utf-8'), replace_all=True)
    if isinstance(output, bytes):
        ans = output  # Assume output is already UTF-8 XML
    elif isinstance(output, str):
        ans = output.encode('utf-8')
    else:
        ans = etree.tostring(output, encoding='utf-8', xml_declaration=True, pretty_print=True)
    return ans


def format_tag_string(tags: list[str] | str, sep: str | None, joinval: str = ', ') -> str:
    if tags:
        tlist = list(tags) if sep is None else [t.strip() for t in cast(str, tags).split(sep)]
    else:
        tlist = []
    tlist.sort(key=sort_key)
    return joinval.join(tlist) if tlist else ''


# Vocabulary for building OPDS feeds {{{
DC_NS: str = 'http://purl.org/dc/terms/'
E: ElementMaker[etree.Element] = ElementMaker(
    namespace='http://www.w3.org/2005/Atom',
    nsmap={
        None: 'http://www.w3.org/2005/Atom',
        'dc': DC_NS,
        'opds': 'http://opds-spec.org/2010/catalog',
    },
)

FEED: Callable[..., etree.Element] = E.feed
TITLE: Callable[..., etree.Element] = E.title
ID: Callable[..., etree.Element] = E.id
ICON: Callable[..., etree.Element] = E.icon


def UPDATED(dt: datetime.datetime, *args: Any, **kwargs: Any) -> etree.Element:
    return E.updated(as_utc(dt).strftime('%Y-%m-%dT%H:%M:%S+00:00'), *args, **kwargs)


@lru_cache(maxsize=2)
def rel_title_map() -> dict[str, str]:
    return {'start': _('Start'), 'up': _('Up'), 'first': _('First'), 'last': _('Last'), 'next': _('Next'), 'previous': _('Previous')}


class LINK:
    type: str = 'application/atom+xml'
    catalog_type: str = ''
    rel: str = ''

    def __call__(self, **kw: str) -> etree.Element:
        t = self.type
        if self.catalog_type:
            t += f';kind={self.catalog_type}'
        kw['type'] = kw.get('type', t)
        if self.rel:
            kw['rel'] = kw.get('rel', self.rel)
            if title := rel_title_map().get(kw['rel'], ''):
                kw['title'] = kw.get('title', title)
        return E.link(**kw)


class NAVLINK(LINK):
    type = LINK.type + ';type=feed;profile=opds-catalog'


class CATALOG_LINK(NAVLINK):
    catalog_type = 'navigation'


def SEARCH_LINK(url_for: Callable[..., str], **kwargs: str) -> etree.Element:
    kwargs['rel'] = 'search'
    kwargs['title'] = 'Search'
    kwargs['href'] = url_for('/opds/search', query='XXX').replace('XXX', '{searchTerms}')
    return LINK()(**kwargs)


def AUTHOR(name: str, uri: str | None = None) -> etree.Element:
    args = [E.name(name)]
    if uri is not None:
        args.append(E.uri(uri))
    return E.author(*args)


SUBTITLE: Callable[..., etree.Element] = E.subtitle


def NAVCATALOG_ENTRY(url_for: UrlForCallable, updated: datetime.datetime, title: str, description: str, query: str) -> etree.Element:
    href = url_for('/opds/navcatalog', which=as_hex_unicode(query))
    id_ = 'calibre-navcatalog:' + hashlib.sha1(as_bytes(href)).hexdigest()
    return E.entry(TITLE(title), ID(id_), UPDATED(updated), E.content(description, type='text'), CATALOG_LINK()(href=href))


class START_LINK(CATALOG_LINK):
    rel = 'start'


class UP_LINK(CATALOG_LINK):
    rel = 'up'


class FIRST_LINK(CATALOG_LINK):
    rel = 'first'


class LAST_LINK(CATALOG_LINK):
    rel = 'last'


class NEXT_LINK(CATALOG_LINK):
    rel = 'next'


class PREVIOUS_LINK(CATALOG_LINK):
    rel = 'previous'


def html_to_lxml(raw: str) -> etree.Element:
    raw = f'<div>{raw}</div>'
    root = parse(raw, keep_doctype=False, namespace_elements=False, maybe_xhtml=False, sanitize_names=True)
    root = next(root.iterdescendants('div'))
    root.set('xmlns', 'http://www.w3.org/1999/xhtml')
    raw = etree.tostring(root, encoding='unicode')
    try:
        return safe_xml_fromstring(raw, recover=False)
    except Exception:
        for x in root.iterdescendants():
            remove = []
            for attr in x.attrib:
                if ':' in attr:
                    remove.append(attr)
            for a in remove:
                del x.attrib[a]
        raw = etree.tostring(root, encoding='unicode')
        try:
            return safe_xml_fromstring(raw, recover=False)
        except Exception:
            from calibre.ebooks.oeb.parse_utils import _html4_parse

            return _html4_parse(raw)


def CATALOG_ENTRY(
    item: Tag,
    item_kind: str,
    request_context: RequestContext,
    updated: datetime.datetime,
    catalog_name: str,
    ignore_count: bool = False,
    add_kind: bool = False,
) -> etree.Element:
    id_ = 'calibre:category:' + item.name
    iid = 'N' + item.name
    if item.id is not None:
        iid = 'I' + str(item.id)
        iid += ':' + item_kind
    href = request_context.url_for('/opds/category', category=as_hex_unicode(catalog_name), which=as_hex_unicode(iid))
    link = CATALOG_LINK()(href=href)
    if ignore_count:
        count = ''
    else:
        count = ngettext('one book', '{} books', item.count).format(item.count)
    if item.use_sort_as_name:
        name = item.sort
    else:
        name = item.name
    return E.entry(
        TITLE(str(name) + ('' if not add_kind else f' ({item_kind})')),
        ID(id_),
        UPDATED(updated),
        E.content(count, type='text'),
        link,
    )


def CATALOG_GROUP_ENTRY(item: Group, category: str, request_context: RequestContext, updated: datetime.datetime) -> etree.Element:
    id_ = 'calibre:category-group:' + category + ':' + (item.text or '')
    iid = item.text
    link = CATALOG_LINK()(href=request_context.url_for('/opds/categorygroup', category=as_hex_unicode(category), which=as_hex_unicode(iid)))
    return E.entry(
        TITLE(item.text),
        ID(id_),
        UPDATED(updated),
        E.content(ngettext('one item', '{} items', item.count).format(item.count), type='text'),
        link,
    )


def ACQUISITION_ENTRY(book_id: int, updated: datetime.datetime, request_context: RequestContext) -> etree.Element:
    field_metadata = request_context.db.field_metadata
    mi = request_context.db.get_metadata(book_id)
    extra = []
    if (mi.rating or 0) > 0:
        rating = rating_to_stars(mi.rating)
        extra.append(_('RATING: %s<br />') % rating)
    if mi.tags:
        extra.append(_('TAGS: %s<br />') % xml(format_tag_string(mi.tags, None)))
    if mi.series:
        extra.append(_('SERIES: %(series)s [%(sidx)s]<br />') % dict(series=xml(mi.series), sidx=fmt_sidx(float(mi.series_index))))
    for key in filter(request_context.ctx.is_field_displayable, field_metadata.ignorable_field_keys()):
        name, val = mi.format_field(key)
        if val:
            fm = field_metadata[key]
            datatype = fm['datatype']
            if datatype == 'text' and fm['is_multiple']:
                extra.append(
                    '{}: {}<br />'.format(
                        xml(name),
                        xml(format_tag_string(val, fm['is_multiple']['ui_to_list'], joinval=fm['is_multiple']['list_to_ui'])),
                    )
                )
            elif datatype == 'comments' or (fm['datatype'] == 'composite' and fm['display'].get('contains_html', False)):
                extra.append(f'{xml(name)}: {comments_to_html(str(val))}<br />')
            else:
                extra.append(f'{xml(name)}: {xml(str(val))}<br />')
    if mi.comments:
        comments = comments_to_html(mi.comments)
        extra.append(comments)
    if extra:
        extra = html_to_lxml('\n'.join(extra))
    ans = E.entry(
        TITLE(mi.title),
        E.author(E.name(authors_to_string(mi.authors))),
        ID('urn:uuid:' + (mi.uuid or '')),
        UPDATED(mi.last_modified),
        E.published(mi.timestamp.isoformat()),
    )
    if mi.pubdate and not is_date_undefined(mi.pubdate):
        ans.append(ans.makeelement(f'{{{DC_NS}}}date'))
        ans[-1].text = mi.pubdate.isoformat()
    if len(extra):
        ans.append(E.content(*extra, type='xhtml'))

    def get(**kwargs: Any) -> str:
        return request_context.ctx.url_for('/get', book_id=book_id, library_id=request_context.library_id, **kwargs)

    if mi.formats:
        fm = mi.format_metadata
        for fmt in mi.formats:
            fmt = fmt.lower()
            mt = guess_type('a.' + fmt)[0]
            if mt:
                link = E.link(type=mt, href=get(what=fmt), rel='http://opds-spec.org/acquisition')
                ffm = fm.get(fmt.upper())
                if ffm:
                    link.set('length', str(ffm['size']))
                    link.set('mtime', ffm['mtime'].isoformat())
                ans.append(link)
    ans.append(E.link(type='image/jpeg', href=get(what='cover'), rel='http://opds-spec.org/cover'))
    ans.append(E.link(type='image/jpeg', href=get(what='thumb'), rel='http://opds-spec.org/thumbnail'))
    ans.append(E.link(type='image/jpeg', href=get(what='cover'), rel='http://opds-spec.org/image'))
    ans.append(E.link(type='image/jpeg', href=get(what='thumb'), rel='http://opds-spec.org/image/thumbnail'))

    return ans


# }}}

default_feed_title: str = __appname__ + ' ' + _('Library')


class Feed:  # {{{
    base_href: str
    root: etree.Element
    is_acquisition_feed: bool = False

    def link(self, link_cls: type[CATALOG_LINK] = CATALOG_LINK, **kw: str) -> etree.Element:
        cl = link_cls()
        if self.is_acquisition_feed:
            cl.catalog_type = 'acquisition'
        return cl(**kw)

    def __init__(
        self,
        id_: str,
        updated: datetime.datetime,
        request_context: RequestContext,
        subtitle: str | None = None,
        title: str | None = None,
        up_link: str | None = None,
        first_link: str | None = None,
        last_link: str | None = None,
        next_link: str | None = None,
        previous_link: str | None = None,
    ) -> None:
        self.base_href = request_context.url_for('/opds')

        self.root = FEED(
            TITLE(title or default_feed_title),
            AUTHOR(__appname__, uri='https://calibre-ebook.com'),
            ID(id_),
            ICON(request_context.ctx.url_for('/favicon.png')),
            UPDATED(updated),
            SEARCH_LINK(request_context.url_for),
            START_LINK()(href=request_context.url_for('/opds')),
        )
        if up_link:
            self.root.append(UP_LINK()(href=up_link))
        if first_link:
            self.root.append(self.link(FIRST_LINK, href=first_link))
        if last_link:
            self.root.append(self.link(LAST_LINK, href=last_link))
        if next_link:
            self.root.append(self.link(NEXT_LINK, href=next_link))
        if previous_link:
            self.root.append(self.link(PREVIOUS_LINK, href=previous_link))
        if subtitle:
            self.root.insert(1, SUBTITLE(subtitle))


# }}}


class TopLevel(Feed):  # {{{
    def __init__(
        self,
        updated: datetime.datetime,  # datetime object in UTC
        categories: list[tuple[str, str, str]],
        request_context: RequestContext,
        id_: str = 'urn:calibre:main',
        subtitle: str = _('Books in your library'),
    ) -> None:
        Feed.__init__(self, id_, updated, request_context, subtitle=subtitle)

        subc = partial(NAVCATALOG_ENTRY, request_context.url_for, updated)
        subcatalogs = [subc(_('By ') + title, _('Books sorted by ') + desc, q) for title, desc, q in categories]
        for x in subcatalogs:
            self.root.append(x)
        for library_id, library_name in sorted(request_context.library_map.items(), key=lambda item: sort_key(item[1])):
            id_ = 'calibre-library:' + library_id
            self.root.append(
                E.entry(
                    TITLE(_('Library:') + ' ' + library_name),
                    ID(id_),
                    UPDATED(updated),
                    E.content(_('Change calibre library to:') + ' ' + library_name, type='text'),
                    self.link(href=request_context.url_for('/opds', library_id=library_id)),
                )
            )


# }}}


class NavFeed(Feed):
    def __init__(
        self,
        id_: str,
        updated: datetime.datetime,
        request_context: RequestContext,
        offsets: Offsets,
        page_url: str,
        up_url: str,
        title: str | None = None,
    ) -> None:
        kwargs: dict[str, Any] = {'up_link': up_url}
        kwargs['first_link'] = page_url
        kwargs['last_link'] = page_url + f'&offset={offsets.last_offset}'
        if offsets.offset > 0:
            kwargs['previous_link'] = page_url + f'&offset={offsets.previous_offset}'
        if offsets.next_offset > -1:
            kwargs['next_link'] = page_url + f'&offset={offsets.next_offset}'
        if title:
            kwargs['title'] = title
        Feed.__init__(self, id_, updated, request_context, **kwargs)


class AcquisitionFeed(NavFeed):
    is_acquisition_feed = True

    def __init__(
        self,
        id_: str,
        updated: datetime.datetime,
        request_context: RequestContext,
        items: list[int],
        offsets: Offsets,
        page_url: str,
        up_url: str,
        title: str | None = None,
    ) -> None:
        NavFeed.__init__(self, id_, updated, request_context, offsets, page_url, up_url, title=title)
        for book_id in items:
            self.root.append(ACQUISITION_ENTRY(book_id, updated, request_context))


class CategoryFeed(NavFeed):
    def __init__(
        self,
        items: list[Tag],
        which: str,
        id_: str,
        updated: datetime.datetime,
        request_context: RequestContext,
        offsets: Offsets,
        page_url: str,
        up_url: str,
        title: str | None = None,
    ) -> None:
        NavFeed.__init__(self, id_, updated, request_context, offsets, page_url, up_url, title=title)
        ignore_count = False
        if which == 'search':
            ignore_count = True
        for item in items:
            self.root.append(
                CATALOG_ENTRY(
                    item,
                    item.category or '',
                    request_context,
                    updated,
                    which,
                    ignore_count=ignore_count,
                    add_kind=which != item.category,
                )
            )


class CategoryGroupFeed(NavFeed):
    def __init__(
        self,
        items: list[Group],
        which: str,
        id_: str,
        updated: datetime.datetime,
        request_context: RequestContext,
        offsets: Offsets,
        page_url: str,
        up_url: str,
        title: str | None = None,
    ) -> None:
        NavFeed.__init__(self, id_, updated, request_context, offsets, page_url, up_url, title=title)
        for item in items:
            self.root.append(CATALOG_GROUP_ENTRY(item, which, request_context, updated))


class RequestContext:
    db: Cache
    library_id: str
    library_map: dict[str, str]
    default_library: str
    ctx: Context
    rd: RequestData

    def __init__(self, ctx: Context, rd: RequestData) -> None:
        self.db, self.library_id, self.library_map, self.default_library = get_library_data(ctx, rd)
        self.ctx, self.rd = ctx, rd

    def url_for(self, route: str | None, **kwargs: Any) -> str:
        lid = kwargs.pop('library_id', self.library_id)
        assert self.ctx.url_for is not None
        ans = self.ctx.url_for(route, **kwargs)
        q = {'library_id': lid}
        ans += '?' + urlencode(q)
        return ans

    def allowed_book_ids(self) -> frozenset[int]:
        return self.ctx.allowed_book_ids(self.rd, self.db)

    @property
    def outheaders(self) -> MultiDict:
        return self.rd.outheaders

    @property
    def opts(self) -> Options:
        return self.ctx.opts

    def last_modified(self) -> datetime.datetime:
        return self.db.last_modified()

    def get_categories(self, report_parse_errors: bool = False) -> dict[str, list[Tag]]:
        return self.ctx.get_categories(self.rd, self.db, report_parse_errors=report_parse_errors)

    def search(self, query: str) -> frozenset[int]:
        return self.ctx.search(self.rd, self.db, query)


def get_acquisition_feed(
    rc: RequestContext,
    ids: frozenset[int],
    offset: int,
    page_url: str,
    up_url: str,
    id_: str,
    sort_by: str = 'title',
    ascending: bool = True,
    feed_title: str | None = None,
) -> etree.Element:
    if not ids:
        raise HTTPNotFound('No books found')
    with rc.db.safe_read_lock:
        sort_by = sanitize_sort_field_name(rc.db.field_metadata, sort_by)
        items = rc.db.multisort([(sort_by, ascending)], ids)
        max_items = rc.opts.max_opds_items
        offsets = Offsets(offset, max_items, len(items))
        items = items[offsets.offset : offsets.offset + max_items]
        lm = rc.last_modified()
        rc.outheaders['Last-Modified'] = http_date(timestampfromdt(lm))
        return AcquisitionFeed(id_, lm, rc, items, offsets, page_url, up_url, title=feed_title).root


def get_all_books(rc: RequestContext, which: str, page_url: str, up_url: str, offset: int = 0) -> etree.Element:
    try:
        offset = int(offset)
    except Exception:
        raise HTTPNotFound('Not found')
    if which not in ('title', 'newest'):
        raise HTTPNotFound('Not found')
    sort = 'timestamp' if which == 'newest' else 'title'
    ascending = which == 'title'
    feed_title = {'newest': _('Newest'), 'title': _('Title')}.get(which, which)
    feed_title = default_feed_title + ' :: ' + _('By %s') % feed_title
    ids = rc.allowed_book_ids()
    return get_acquisition_feed(
        rc,
        ids,
        offset,
        page_url,
        up_url,
        id_='calibre-all:' + sort,
        sort_by=sort,
        ascending=ascending,
        feed_title=feed_title,
    )


def get_navcatalog(request_context: RequestContext, which: str, page_url: str, up_url: str, offset: int = 0) -> etree.Element:
    categories = request_context.get_categories()
    if which not in categories:
        raise HTTPNotFound(f'Category {which!r} not found')

    items = categories[which]
    updated = request_context.last_modified()
    category_meta = request_context.db.field_metadata
    meta = category_meta.get(which, {})
    category_name = meta.get('name', which)
    feed_title = default_feed_title + ' :: ' + _('By %s') % category_name

    id_ = 'calibre-category-feed:' + which

    MAX_ITEMS = request_context.opts.max_opds_ungrouped_items

    if MAX_ITEMS > 0 and len(items) <= MAX_ITEMS:
        max_items = request_context.opts.max_opds_items
        offsets = Offsets(offset, max_items, len(items))
        items = list(items)[offsets.offset : offsets.offset + max_items]
        ans = CategoryFeed(items, which, id_, updated, request_context, offsets, page_url, up_url, title=feed_title)
    else:
        starts = set()
        for x in items:
            val = getattr(x, 'sort', x.name)
            if not val:
                val = 'A'
            starts.add(val[0].upper())
        category_groups = OrderedDict()
        for x in sorted(starts, key=sort_key):
            category_groups[x] = len([y for y in items if getattr(y, 'sort', y.name).upper().startswith(x)])
        items = [Group(x, y) for x, y in category_groups.items()]
        max_items = request_context.opts.max_opds_items
        offsets = Offsets(offset, max_items, len(items))
        items = items[offsets.offset : offsets.offset + max_items]
        ans = CategoryGroupFeed(items, which, id_, updated, request_context, offsets, page_url, up_url, title=feed_title)

    request_context.outheaders['Last-Modified'] = http_date(timestampfromdt(updated))

    return ans.root


@endpoint('/opds', postprocess=atom)
def opds(ctx: Context, rd: RequestData) -> etree.Element:
    rc = RequestContext(ctx, rd)
    db = rc.db
    try:
        categories = rc.get_categories(report_parse_errors=True)
    except ParseException as p:
        raise HTTPInternalServerError(p.msg)

    category_meta = db.field_metadata
    cats = [
        (_('Newest'), _('Date'), 'Onewest'),
        (_('Title'), _('Title'), 'Otitle'),
    ]

    def getter(x: str) -> str:
        try:
            return category_meta[x]['name'].lower()
        except KeyError:
            return x

    fm = rc.db.field_metadata
    for category in sorted(categories, key=lambda x: sort_key(getter(x))):
        if fm.is_ignorable_field(category) and not rc.ctx.is_field_displayable(category):
            continue
        if len(categories[category]) == 0:
            continue
        if category in ('formats', 'identifiers'):
            continue
        meta = category_meta.get(category, None)
        if meta is None:
            continue
        cats.append((meta['name'], meta['name'], 'N' + category))
    last_modified = db.last_modified()
    rd.outheaders['Last-Modified'] = http_date(timestampfromdt(last_modified))
    return TopLevel(last_modified, cats, rc).root


@endpoint('/opds/navcatalog/{which}', postprocess=atom)
def opds_navcatalog(ctx: Context, rd: RequestData, which: str) -> etree.Element:
    try:
        offset = int(rd.query.get('offset', 0))
    except Exception:
        raise HTTPNotFound('Not found')
    rc = RequestContext(ctx, rd)

    page_url = rc.url_for('/opds/navcatalog', which=which)
    up_url = rc.url_for('/opds')
    which = from_hex_unicode(which)
    type_ = which[0]
    which = which[1:]
    if type_ == 'O':
        return get_all_books(rc, which, page_url, up_url, offset=offset)
    elif type_ == 'N':
        return get_navcatalog(rc, which, page_url, up_url, offset=offset)
    raise HTTPNotFound('Not found')


@endpoint('/opds/category/{category}/{which}', postprocess=atom)
def opds_category(ctx: Context, rd: RequestData, category: str, which: str) -> etree.Element:
    try:
        offset = int(rd.query.get('offset', 0))
    except Exception:
        raise HTTPNotFound('Not found')

    if not which or not category:
        raise HTTPNotFound('Not found')
    rc = RequestContext(ctx, rd)
    page_url = rc.url_for('/opds/category', which=which, category=category)
    up_url = rc.url_for('/opds/navcatalog', which=category)

    which, category = from_hex_unicode(which), from_hex_unicode(category)
    type_ = which[0]
    which = which[1:]
    if type_ == 'I':
        try:
            p = which.rindex(':')
            category = which[p + 1 :]
            which = which[:p]
            # This line will toss an exception for composite columns
            which = str(int(which[:p]))
        except Exception:
            # Might be a composite column, where we have the lookup key
            if not (category in rc.db.field_metadata and rc.db.field_metadata[category]['datatype'] == 'composite'):
                raise HTTPNotFound(f'Tag {which!r} not found')

    categories = rc.get_categories()
    if category not in categories:
        raise HTTPNotFound(f'Category {which!r} not found')

    if category == 'search':
        try:
            ids = rc.search(f'search:"{which}"')
        except Exception:
            raise HTTPNotFound(f'Search: {which!r} not understood')
        return get_acquisition_feed(rc, ids, offset, page_url, up_url, 'calibre-search:' + which)

    if type_ != 'I':
        raise HTTPNotFound('Non id categories not supported')

    q = category
    if q == 'news':
        q = 'tags'
    ids = rc.db.get_books_for_category(q, which) & rc.allowed_book_ids()
    sort_by = 'series' if category == 'series' else 'title'

    return get_acquisition_feed(rc, ids, offset, page_url, up_url, 'calibre-category:' + category + ':' + str(which), sort_by=sort_by)


@endpoint('/opds/categorygroup/{category}/{which}', postprocess=atom)
def opds_categorygroup(ctx: Context, rd: RequestData, category: str, which: str) -> etree.Element:
    try:
        offset = int(rd.query.get('offset', 0))
    except Exception:
        raise HTTPNotFound('Not found')

    if not which or not category:
        raise HTTPNotFound('Not found')

    rc = RequestContext(ctx, rd)
    categories = rc.get_categories()
    page_url = rc.url_for('/opds/categorygroup', category=category, which=which)

    category = from_hex_unicode(category)
    if category not in categories:
        raise HTTPNotFound(f'Category {which!r} not found')
    category_meta = rc.db.field_metadata
    meta = category_meta.get(category, {})
    category_name = meta.get('name', which)
    which = from_hex_unicode(which)
    feed_title = default_feed_title + ' :: ' + (_('By {0} :: {1}').format(category_name, which))
    owhich = as_hex_unicode('N' + category)
    up_url = rc.url_for('/opds/navcatalog', which=owhich)
    items = categories[category]

    def belongs(x: Tag, which: str) -> bool:
        return getattr(x, 'sort', x.name).lower().startswith(which.lower())

    items = [x for x in items if belongs(x, which)]
    if not items:
        raise HTTPNotFound(f'No items in group {category!r}:{which!r}')
    updated = rc.last_modified()

    id_ = 'calibre-category-group-feed:' + category + ':' + which

    max_items = rc.opts.max_opds_items
    offsets = Offsets(offset, max_items, len(items))
    items = list(items)[offsets.offset : offsets.offset + max_items]

    rc.outheaders['Last-Modified'] = http_date(timestampfromdt(updated))

    return CategoryFeed(items, category, id_, updated, rc, offsets, page_url, up_url, title=feed_title).root


@endpoint('/opds/search/{query=""}', postprocess=atom)
def opds_search(ctx: Context, rd: RequestData, query: str) -> etree.Element:
    try:
        offset = int(rd.query.get('offset', 0))
    except Exception:
        raise HTTPNotFound('Not found')

    rc = RequestContext(ctx, rd)
    if query:
        path = parse_uri(rd.request_original_uri, parse_query=False, unquote_func=unquote_plus)[1]
        query = path[-1]
        if isinstance(query, bytes):
            query = query.decode('utf-8')
    try:
        ids = rc.search(query)
    except Exception:
        raise HTTPNotFound(f'Search: {query!r} not understood')
    page_url = rc.url_for('/opds/search', query=query)
    return get_acquisition_feed(rc, ids, offset, page_url, rc.url_for('/opds'), 'calibre-search:' + query)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

from calibre.ebooks.conversion.config import OPTIONS
from calibre.gui2.convert import Widget
from calibre.gui2.convert.epub_output_ui import Ui_Form
from calibre.utils.localization import _


class PluginWidget(Widget, Ui_Form):
    TITLE = _('EPUB output')
    HELP = _('Options specific to') + ' EPUB ' + _('output')
    COMMIT_NAME = 'epub_output'
    ICON = 'mimetypes/epub.png'

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent, OPTIONS['output']['epub'])
        for i in range(2):
            self.opt_no_svg_cover.toggle()
        ev = get_option('epub_version')
        self.opt_epub_version.addItems(list(ev.option.choices))
        self.db, self.book_id = db, book_id
        self.initialize_options(get_option, get_help, db, book_id)



# -*- coding: utf-8 -*-
# License: GPLv3 Copyright: 2012-2023, Tomasz Długosz <tomek3d@gmail.com>

from __future__ import absolute_import, division, print_function, unicode_literals

store_version = 6  # Needed for dynamic plugin loading

from contextlib import closing
from urllib.parse import quote_plus

from qt.core import QUrl

from calibre import browser, url_slash_cleaner
from calibre.gui2 import open_url
from calibre.gui2.store import StorePlugin
from calibre.gui2.store.basic_config import BasicStoreConfig
from calibre.gui2.store.search_result import SearchResult
from calibre.gui2.store.web_store_dialog import WebStoreDialog

try:
    from calibre.utils.xml_parse import safe_html_fromstring
except ImportError:
    from lxml.html import fromstring as safe_html_fromstring

import json


class WolneLekturyStore(BasicStoreConfig, StorePlugin):
    def open(self, gui=None, parent=None, detail_item=None, external=False):

        url = 'https://wolnelektury.pl'
        detail_url = None

        if detail_item:
            detail_url = detail_item

        if external or self.config.get('open_external', False):
            open_url(QUrl(url_slash_cleaner(detail_url or url)))
        else:
            d = WebStoreDialog(self.gui, url, parent, detail_url)
            d.setWindowTitle(self.name)
            d.set_tags(self.config.get('tags', ''))
            d.exec()

    def search(self, query, max_results=10, timeout=60):
        url = 'https://wolnelektury.pl/szukaj?q=' + quote_plus(query)

        br = browser()
        price = '0,00 zł'

        counter = max_results
        with closing(br.open(url, timeout=timeout)) as f:
            doc = safe_html_fromstring(f.read())
            for data in doc.xpath('//div[@class="l-books__grid"]/article'):
                if counter <= 0:
                    break

                try:
                    id = ''.join(data.xpath('.//figure/a/@href')).split('/')[3]
                except IndexError:
                    continue

                title = ''.join(data.xpath('.//h2/a/text()'))
                author = ', '.join(data.xpath('.//h3/a/text()'))
                cover_url = ''.join(data.xpath('.//figure/a/img/@src'))

                s = SearchResult()
                s.cover_url = 'https://wolnelektury.pl' + cover_url.strip()
                s.title = title.strip()
                s.author = author
                s.price = price
                s.detail_item = 'https://wolnelektury.pl/katalog/lektura/' + id

                s.downloads.update(self._search_formats(id, timeout=timeout))

                s.formats = ', '.join(s.downloads.keys())
                s.drm = SearchResult.DRM_UNLOCKED

                counter -= 1

                yield s

    def _search_formats(self, id: str, timeout: int = 60) -> dict[str, str]:
        # formats used by the site and calibre (as of 01.05.2026)
        ALLOWED_FORMATS: tuple[str, ...] = ('pdf', 'epub', 'mobi', 'txt', 'html', 'fb2')
        result: dict[str, str] = {}
        br = browser()
        url = f'https://wolnelektury.pl/api/books/{id}/?format=json'
        with closing(br.open(url, timeout=timeout)) as page:
            parsed_data = json.load(page)
            for ext in ALLOWED_FORMATS:
                if (book_url := parsed_data.get(ext)) is None:
                    continue
                if book_url != '':
                    result[ext] = book_url

        return result



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

# cff_standard_strings {{{
# The 391 Standard Strings as used in the CFF format.
# from Adobe Technical None #5176, version 1.0, 18 March 1998

cff_standard_strings = [
    '.notdef',
    'space',
    'exclam',
    'quotedbl',
    'numbersign',
    'dollar',
    'percent',
    'ampersand',
    'quoteright',
    'parenleft',
    'parenright',
    'asterisk',
    'plus',
    'comma',
    'hyphen',
    'period',
    'slash',
    'zero',
    'one',
    'two',
    'three',
    'four',
    'five',
    'six',
    'seven',
    'eight',
    'nine',
    'colon',
    'semicolon',
    'less',
    'equal',
    'greater',
    'question',
    'at',
    'A',
    'B',
    'C',
    'D',
    'E',
    'F',
    'G',
    'H',
    'I',
    'J',
    'K',
    'L',
    'M',
    'N',
    'O',
    'P',
    'Q',
    'R',
    'S',
    'T',
    'U',
    'V',
    'W',
    'X',
    'Y',
    'Z',
    'bracketleft',
    'backslash',
    'bracketright',
    'asciicircum',
    'underscore',
    'quoteleft',
    'a',
    'b',
    'c',
    'd',
    'e',
    'f',
    'g',
    'h',
    'i',
    'j',
    'k',
    'l',
    'm',
    'n',
    'o',
    'p',
    'q',
    'r',
    's',
    't',
    'u',
    'v',
    'w',
    'x',
    'y',
    'z',
    'braceleft',
    'bar',
    'braceright',
    'asciitilde',
    'exclamdown',
    'cent',
    'sterling',
    'fraction',
    'yen',
    'florin',
    'section',
    'currency',
    'quotesingle',
    'quotedblleft',
    'guillemotleft',
    'guilsinglleft',
    'guilsinglright',
    'fi',
    'fl',
    'endash',
    'dagger',
    'daggerdbl',
    'periodcentered',
    'paragraph',
    'bullet',
    'quotesinglbase',
    'quotedblbase',
    'quotedblright',
    'guillemotright',
    'ellipsis',
    'perthousand',
    'questiondown',
    'grave',
    'acute',
    'circumflex',
    'tilde',
    'macron',
    'breve',
    'dotaccent',
    'dieresis',
    'ring',
    'cedilla',
    'hungarumlaut',
    'ogonek',
    'caron',
    'emdash',
    'AE',
    'ordfeminine',
    'Lslash',
    'Oslash',
    'OE',
    'ordmasculine',
    'ae',
    'dotlessi',
    'lslash',
    'oslash',
    'oe',
    'germandbls',
    'onesuperior',
    'logicalnot',
    'mu',
    'trademark',
    'Eth',
    'onehalf',
    'plusminus',
    'Thorn',
    'onequarter',
    'divide',
    'brokenbar',
    'degree',
    'thorn',
    'threequarters',
    'twosuperior',
    'registered',
    'minus',
    'eth',
    'multiply',
    'threesuperior',
    'copyright',
    'Aacute',
    'Acircumflex',
    'Adieresis',
    'Agrave',
    'Aring',
    'Atilde',
    'Ccedilla',
    'Eacute',
    'Ecircumflex',
    'Edieresis',
    'Egrave',
    'Iacute',
    'Icircumflex',
    'Idieresis',
    'Igrave',
    'Ntilde',
    'Oacute',
    'Ocircumflex',
    'Odieresis',
    'Ograve',
    'Otilde',
    'Scaron',
    'Uacute',
    'Ucircumflex',
    'Udieresis',
    'Ugrave',
    'Yacute',
    'Ydieresis',
    'Zcaron',
    'aacute',
    'acircumflex',
    'adieresis',
    'agrave',
    'aring',
    'atilde',
    'ccedilla',
    'eacute',
    'ecircumflex',
    'edieresis',
    'egrave',
    'iacute',
    'icircumflex',
    'idieresis',
    'igrave',
    'ntilde',
    'oacute',
    'ocircumflex',
    'odieresis',
    'ograve',
    'otilde',
    'scaron',
    'uacute',
    'ucircumflex',
    'udieresis',
    'ugrave',
    'yacute',
    'ydieresis',
    'zcaron',
    'exclamsmall',
    'Hungarumlautsmall',
    'dollaroldstyle',
    'dollarsuperior',
    'ampersandsmall',
    'Acutesmall',
    'parenleftsuperior',
    'parenrightsuperior',
    'twodotenleader',
    'onedotenleader',
    'zerooldstyle',
    'oneoldstyle',
    'twooldstyle',
    'threeoldstyle',
    'fouroldstyle',
    'fiveoldstyle',
    'sixoldstyle',
    'sevenoldstyle',
    'eightoldstyle',
    'nineoldstyle',
    'commasuperior',
    'threequartersemdash',
    'periodsuperior',
    'questionsmall',
    'asuperior',
    'bsuperior',
    'centsuperior',
    'dsuperior',
    'esuperior',
    'isuperior',
    'lsuperior',
    'msuperior',
    'nsuperior',
    'osuperior',
    'rsuperior',
    'ssuperior',
    'tsuperior',
    'ff',
    'ffi',
    'ffl',
    'parenleftinferior',
    'parenrightinferior',
    'Circumflexsmall',
    'hyphensuperior',
    'Gravesmall',
    'Asmall',
    'Bsmall',
    'Csmall',
    'Dsmall',
    'Esmall',
    'Fsmall',
    'Gsmall',
    'Hsmall',
    'Ismall',
    'Jsmall',
    'Ksmall',
    'Lsmall',
    'Msmall',
    'Nsmall',
    'Osmall',
    'Psmall',
    'Qsmall',
    'Rsmall',
    'Ssmall',
    'Tsmall',
    'Usmall',
    'Vsmall',
    'Wsmall',
    'Xsmall',
    'Ysmall',
    'Zsmall',
    'colonmonetary',
    'onefitted',
    'rupiah',
    'Tildesmall',
    'exclamdownsmall',
    'centoldstyle',
    'Lslashsmall',
    'Scaronsmall',
    'Zcaronsmall',
    'Dieresissmall',
    'Brevesmall',
    'Caronsmall',
    'Dotaccentsmall',
    'Macronsmall',
    'figuredash',
    'hypheninferior',
    'Ogoneksmall',
    'Ringsmall',
    'Cedillasmall',
    'questiondownsmall',
    'oneeighth',
    'threeeighths',
    'fiveeighths',
    'seveneighths',
    'onethird',
    'twothirds',
    'zerosuperior',
    'foursuperior',
    'fivesuperior',
    'sixsuperior',
    'sevensuperior',
    'eightsuperior',
    'ninesuperior',
    'zeroinferior',
    'oneinferior',
    'twoinferior',
    'threeinferior',
    'fourinferior',
    'fiveinferior',
    'sixinferior',
    'seveninferior',
    'eightinferior',
    'nineinferior',
    'centinferior',
    'dollarinferior',
    'periodinferior',
    'commainferior',
    'Agravesmall',
    'Aacutesmall',
    'Acircumflexsmall',
    'Atildesmall',
    'Adieresissmall',
    'Aringsmall',
    'AEsmall',
    'Ccedillasmall',
    'Egravesmall',
    'Eacutesmall',
    'Ecircumflexsmall',
    'Edieresissmall',
    'Igravesmall',
    'Iacutesmall',
    'Icircumflexsmall',
    'Idieresissmall',
    'Ethsmall',
    'Ntildesmall',
    'Ogravesmall',
    'Oacutesmall',
    'Ocircumflexsmall',
    'Otildesmall',
    'Odieresissmall',
    'OEsmall',
    'Oslashsmall',
    'Ugravesmall',
    'Uacutesmall',
    'Ucircumflexsmall',
    'Udieresissmall',
    'Yacutesmall',
    'Thornsmall',
    'Ydieresissmall',
    '001.000',
    '001.001',
    '001.002',
    '001.003',
    'Black',
    'Bold',
    'Book',
    'Light',
    'Medium',
    'Regular',
    'Roman',
    'Semibold',
]
# }}}

STANDARD_CHARSETS = [  # {{{
    # ISOAdobe
    (
        '.notdef',
        'space',
        'exclam',
        'quotedbl',
        'numbersign',
        'dollar',
        'percent',
        'ampersand',
        'quoteright',
        'parenleft',
        'parenright',
        'asterisk',
        'plus',
        'comma',
        'hyphen',
        'period',
        'slash',
        'zero',
        'one',
        'two',
        'three',
        'four',
        'five',
        'six',
        'seven',
        'eight',
        'nine',
        'colon',
        'semicolon',
        'less',
        'equal',
        'greater',
        'question',
        'at',
        'A',
        'B',
        'C',
        'D',
        'E',
        'F',
        'G',
        'H',
        'I',
        'J',
        'K',
        'L',
        'M',
        'N',
        'O',
        'P',
        'Q',
        'R',
        'S',
        'T',
        'U',
        'V',
        'W',
        'X',
        'Y',
        'Z',
        'bracketleft',
        'backslash',
        'bracketright',
        'asciicircum',
        'underscore',
        'quoteleft',
        'a',
        'b',
        'c',
        'd',
        'e',
        'f',
        'g',
        'h',
        'i',
        'j',
        'k',
        'l',
        'm',
        'n',
        'o',
        'p',
        'q',
        'r',
        's',
        't',
        'u',
        'v',
        'w',
        'x',
        'y',
        'z',
        'braceleft',
        'bar',
        'braceright',
        'asciitilde',
        'exclamdown',
        'cent',
        'sterling',
        'fraction',
        'yen',
        'florin',
        'section',
        'currency',
        'quotesingle',
        'quotedblleft',
        'guillemotleft',
        'guilsinglleft',
        'guilsinglright',
        'fi',
        'fl',
        'endash',
        'dagger',
        'daggerdbl',
        'periodcentered',
        'paragraph',
        'bullet',
        'quotesinglbase',
        'quotedblbase',
        'quotedblright',
        'guillemotright',
        'ellipsis',
        'perthousand',
        'questiondown',
        'grave',
        'acute',
        'circumflex',
        'tilde',
        'macron',
        'breve',
        'dotaccent',
        'dieresis',
        'ring',
        'cedilla',
        'hungarumlaut',
        'ogonek',
        'caron',
        'emdash',
        'AE',
        'ordfeminine',
        'Lslash',
        'Oslash',
        'OE',
        'ordmasculine',
        'ae',
        'dotlessi',
        'lslash',
        'oslash',
        'oe',
        'germandbls',
        'onesuperior',
        'logicalnot',
        'mu',
        'trademark',
        'Eth',
        'onehalf',
        'plusminus',
        'Thorn',
        'onequarter',
        'divide',
        'brokenbar',
        'degree',
        'thorn',
        'threequarters',
        'twosuperior',
        'registered',
        'minus',
        'eth',
        'multiply',
        'threesuperior',
        'copyright',
        'Aacute',
        'Acircumflex',
        'Adieresis',
        'Agrave',
        'Aring',
        'Atilde',
        'Ccedilla',
        'Eacute',
        'Ecircumflex',
        'Edieresis',
        'Egrave',
        'Iacute',
        'Icircumflex',
        'Idieresis',
        'Igrave',
        'Ntilde',
        'Oacute',
        'Ocircumflex',
        'Odieresis',
        'Ograve',
        'Otilde',
        'Scaron',
        'Uacute',
        'Ucircumflex',
        'Udieresis',
        'Ugrave',
        'Yacute',
        'Ydieresis',
        'Zcaron',
        'aacute',
        'acircumflex',
        'adieresis',
        'agrave',
        'aring',
        'atilde',
        'ccedilla',
        'eacute',
        'ecircumflex',
        'edieresis',
        'egrave',
        'iacute',
        'icircumflex',
        'idieresis',
        'igrave',
        'ntilde',
        'oacute',
        'ocircumflex',
        'odieresis',
        'ograve',
        'otilde',
        'scaron',
        'uacute',
        'ucircumflex',
        'udieresis',
        'ugrave',
        'yacute',
        'ydieresis',
        'zcaron',
    ),
    # Expert
    (
        'notdef',
        'space',
        'exclamsmall',
        'Hungarumlautsmall',
        'dollaroldstyle',
        'dollarsuperior',
        'ampersandsmall',
        'Acutesmall',
        'parenleftsuperior',
        'parenrightsuperior',
        'twodotenleader',
        'onedotenleader',
        'comma',
        'hyphen',
        'period',
        'fraction',
        'zerooldstyle',
        'oneoldstyle',
        'twooldstyle',
        'threeoldstyle',
        'fouroldstyle',
        'fiveoldstyle',
        'sixoldstyle',
        'sevenoldstyle',
        'eightoldstyle',
        'nineoldstyle',
        'colon',
        'semicolon',
        'commasuperior',
        'threequartersemdash',
        'periodsuperior',
        'questionsmall',
        'asuperior',
        'bsuperior',
        'centsuperior',
        'dsuperior',
        'esuperior',
        'isuperior',
        'lsuperior',
        'msuperior',
        'nsuperior',
        'osuperior',
        'rsuperior',
        'ssuperior',
        'tsuperior',
        'ff',
        'fi',
        'fl',
        'ffi',
        'ffl',
        'parenleftinferior',
        'parenrightinferior',
        'Circumflexsmall',
        'hyphensuperior',
        'Gravesmall',
        'Asmall',
        'Bsmall',
        'Csmall',
        'Dsmall',
        'Esmall',
        'Fsmall',
        'Gsmall',
        'Hsmall',
        'Ismall',
        'Jsmall',
        'Ksmall',
        'Lsmall',
        'Msmall',
        'Nsmall',
        'Osmall',
        'Psmall',
        'Qsmall',
        'Rsmall',
        'Ssmall',
        'Tsmall',
        'Usmall',
        'Vsmall',
        'Wsmall',
        'Xsmall',
        'Ysmall',
        'Zsmall',
        'colonmonetary',
        'onefitted',
        'rupiah',
        'Tildesmall',
        'exclamdownsmall',
        'centoldstyle',
        'Lslashsmall',
        'Scaronsmall',
        'Zcaronsmall',
        'Dieresissmall',
        'Brevesmall',
        'Caronsmall',
        'Dotaccentsmall',
        'Macronsmall',
        'figuredash',
        'hypheninferior',
        'Ogoneksmall',
        'Ringsmall',
        'Cedillasmall',
        'onequarter',
        'onehalf',
        'threequarters',
        'questiondownsmall',
        'oneeighth',
        'threeeighths',
        'fiveeighths',
        'seveneighths',
        'onethird',
        'twothirds',
        'zerosuperior',
        'onesuperior',
        'twosuperior',
        'threesuperior',
        'foursuperior',
        'fivesuperior',
        'sixsuperior',
        'sevensuperior',
        'eightsuperior',
        'ninesuperior',
        'zeroinferior',
        'oneinferior',
        'twoinferior',
        'threeinferior',
        'fourinferior',
        'fiveinferior',
        'sixinferior',
        'seveninferior',
        'eightinferior',
        'nineinferior',
        'centinferior',
        'dollarinferior',
        'periodinferior',
        'commainferior',
        'Agravesmall',
        'Aacutesmall',
        'Acircumflexsmall',
        'Atildesmall',
        'Adieresissmall',
        'Aringsmall',
        'AEsmall',
        'Ccedillasmall',
        'Egravesmall',
        'Eacutesmall',
        'Ecircumflexsmall',
        'Edieresissmall',
        'Igravesmall',
        'Iacutesmall',
        'Icircumflexsmall',
        'Idieresissmall',
        'Ethsmall',
        'Ntildesmall',
        'Ogravesmall',
        'Oacutesmall',
        'Ocircumflexsmall',
        'Otildesmall',
        'Odieresissmall',
        'OEsmall',
        'Oslashsmall',
        'Ugravesmall',
        'Uacutesmall',
        'Ucircumflexsmall',
        'Udieresissmall',
        'Yacutesmall',
        'Thornsmall',
        'Ydieresissmall',
    ),
    # Expert Subset
    (
        '.notdef',
        'space',
        'dollaroldstyle',
        'dollarsuperior',
        'parenleftsuperior',
        'parenrightsuperior',
        'twodotenleader',
        'onedotenleader',
        'comma',
        'hyphen',
        'period',
        'fraction',
        'zerooldstyle',
        'oneoldstyle',
        'twooldstyle',
        'threeoldstyle',
        'fouroldstyle',
        'fiveoldstyle',
        'sixoldstyle',
        'sevenoldstyle',
        'eightoldstyle',
        'nineoldstyle',
        'colon',
        'semicolon',
        'commasuperior',
        'threequartersemdash',
        'periodsuperior',
        'asuperior',
        'bsuperior',
        'centsuperior',
        'dsuperior',
        'esuperior',
        'isuperior',
        'lsuperior',
        'msuperior',
        'nsuperior',
        'osuperior',
        'rsuperior',
        'ssuperior',
        'tsuperior',
        'ff',
        'fi',
        'fl',
        'ffi',
        'ffl',
        'parenleftinferior',
        'parenrightinferior',
        'hyphensuperior',
        'colonmonetary',
        'onefitted',
        'rupiah',
        'centoldstyle',
        'figuredash',
        'hypheninferior',
        'onequarter',
        'onehalf',
        'threequarters',
        'oneeighth',
        'threeeighths',
        'fiveeighths',
        'seveneighths',
        'onethird',
        'twothirds',
        'zerosuperior',
        'onesuperior',
        'twosuperior',
        'threesuperior',
        'foursuperior',
        'fivesuperior',
        'sixsuperior',
        'sevensuperior',
        'eightsuperior',
        'ninesuperior',
        'zeroinferior',
        'oneinferior',
        'twoinferior',
        'threeinferior',
        'fourinferior',
        'fiveinferior',
        'sixinferior',
        'seveninferior',
        'eightinferior',
        'nineinferior',
        'centinferior',
        'dollarinferior',
        'periodinferior',
        'commainferior',
    ),
]  # }}}



# License: GPLv3 Copyright: 2008, Kovid Goyal <kovid at kovidgoyal.net>

"""
This package contains logic to read and write LRF files.
The LRF file format is documented at U{http://www.sven.de/librie/Librie/LrfFormat}.
"""

from calibre.ebooks import ConversionError
from calibre.ebooks.lrf.fonts import FONT_FILE_MAP
from calibre.ebooks.lrf.pylrs.pylrs import BlockStyle, Header, TextBlock, TextStyle
from calibre.ebooks.lrf.pylrs.pylrs import Book as _Book


class LRFParseError(Exception):
    pass


class PRS500_PROFILE:
    screen_width = 600
    screen_height = 775
    dpi = 166
    # Number of pixels to subtract from screen_height when calculating height of text area
    fudge = 0
    font_size = 10  #: Default (in pt)
    parindent = 10  #: Default (in pt)
    line_space = 1.2  # : Default (in pt)
    header_font_size = 6  #: In pt
    header_height = 30  # : In px
    default_fonts = {'sans': 'Swis721 BT Roman', 'mono': 'Courier10 BT Roman', 'serif': 'Dutch801 Rm BT Roman'}

    name = 'prs500'


def find_custom_fonts(options, logger):
    from calibre.utils.fonts.scanner import font_scanner

    fonts = {'serif': None, 'sans': None, 'mono': None}

    def family(cmd):
        return cmd.split(',')[-1].strip()

    if options.serif_family:
        f = family(options.serif_family)
        fonts['serif'] = font_scanner.legacy_fonts_for_family(f)
        if not fonts['serif']:
            logger.warn(f'Unable to find serif family {f}')
    if options.sans_family:
        f = family(options.sans_family)
        fonts['sans'] = font_scanner.legacy_fonts_for_family(f)
        if not fonts['sans']:
            logger.warn(f'Unable to find sans family {f}')
    if options.mono_family:
        f = family(options.mono_family)
        fonts['mono'] = font_scanner.legacy_fonts_for_family(f)
        if not fonts['mono']:
            logger.warn(f'Unable to find mono family {f}')
    return fonts


def Book(options, logger, font_delta=0, header=None, profile=PRS500_PROFILE, **settings):
    from uuid import uuid4

    ps = {}
    ps['topmargin'] = options.top_margin
    ps['evensidemargin'] = options.left_margin
    ps['oddsidemargin'] = options.left_margin
    ps['textwidth'] = profile.screen_width - (options.left_margin + options.right_margin)
    ps['textheight'] = profile.screen_height - (options.top_margin + options.bottom_margin) - profile.fudge
    if header:
        hdr = Header()
        hb = TextBlock(
            textStyle=TextStyle(align='foot', fontsize=int(profile.header_font_size * 10)),
            blockStyle=BlockStyle(blockwidth=ps['textwidth']),
        )
        hb.append(header)
        hdr.PutObj(hb)
        ps['headheight'] = profile.header_height
        ps['headsep'] = options.header_separation
        ps['header'] = hdr
        ps['topmargin'] = 0
        ps['textheight'] = profile.screen_height - (options.bottom_margin + ps['topmargin']) - ps['headheight'] - ps['headsep'] - profile.fudge

    fontsize = int(10 * profile.font_size + font_delta * 20)
    baselineskip = fontsize + 20
    fonts = find_custom_fonts(options, logger)
    tsd = dict(
        fontsize=fontsize,
        parindent=int(10 * profile.parindent),
        linespace=int(10 * profile.line_space),
        baselineskip=baselineskip,
        wordspace=10 * options.wordspace,
    )
    if fonts['serif'] and 'normal' in fonts['serif']:
        tsd['fontfacename'] = fonts['serif']['normal'][1]

    book = _Book(
        textstyledefault=tsd,
        pagestyledefault=ps,
        blockstyledefault=dict(blockwidth=ps['textwidth']),
        bookid=uuid4().hex,
        **settings,
    )
    for family in fonts.keys():
        if fonts[family]:
            for font in fonts[family].values():
                book.embed_font(*font)
                FONT_FILE_MAP[font[1]] = font[0]

    for family in ['serif', 'sans', 'mono']:
        if not fonts[family]:
            fonts[family] = {'normal': (None, profile.default_fonts[family])}
        elif 'normal' not in fonts[family]:
            raise ConversionError('Could not find the normal version of the ' + family + ' font')
    return book, fonts



# License: GPLv3 Copyright: 2009, Kovid Goyal

"""
Device driver for Barns and Nobel's Nook
"""

from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _


class EDGE(USBMS):
    name = 'Edge Device Interface'
    gui_name = _('Entourage Edge')
    description = _('Communicate with the Entourage Edge.')
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'linux', 'osx']

    # Ordered list of supported formats
    FORMATS = ['epub', 'pdf']

    VENDOR_ID = [0x0BB4]
    PRODUCT_ID = [0x0C02]
    BCD = [0x0223]

    VENDOR_NAME = ['ANDROID', 'LINUX']
    WINDOWS_MAIN_MEM = ['__FILE-STOR_GADG', 'FILE-CD_GADGET']
    WINDOWS_CARD_A_MEM = ['__FILE-STOR_GADG', 'FILE-CD_GADGET']

    MAIN_MEMORY_VOLUME_LABEL = 'Edge Main Memory'
    STORAGE_CARD_VOLUME_LABEL = 'Edge Storage Card'

    EBOOK_DIR_MAIN = 'download'
    SUPPORTS_SUB_DIRS = True



#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

import errno
import io
import json
import os
import re
import subprocess
import sys
import tempfile
import time
from base64 import standard_b64decode, standard_b64encode
from functools import lru_cache
from typing import Any

from calibre.constants import (
    FAKE_HOST,
    FAKE_PROTOCOL,
    SPECIAL_TITLE_FOR_WEBENGINE_COMMS,
    __appname__,
    __version__,
    builtin_colors_dark,
    builtin_colors_light,
    builtin_decorations,
    dark_link_color,
)
from calibre.utils.filenames import atomic_rename, is_existing_subpath, is_path_inside
from calibre.utils.resources import get_path as P
from polyglot.builtins import as_bytes, exec_path

COMPILER_PATH = 'rapydscript/compiler.js.xz'


def abspath(x):
    return os.path.realpath(os.path.abspath(x))


# Update RapydScript {{{


def update_rapydscript():
    import lzma

    with tempfile.TemporaryDirectory() as tdir:
        subprocess.check_call(['rapydscript', 'web-repl-export', tdir])
        with open(os.path.join(tdir, 'rapydscript.js'), 'rb') as f:
            raw = f.read()
    path = P(COMPILER_PATH, allow_user_override=False)
    with lzma.open(path, 'wb', format=lzma.FORMAT_XZ) as f:
        f.write(raw)


# }}}

# Compiler {{{


def to_dict(obj):
    return dict(zip(list(obj.keys()), list(obj.values())))


_compiler_instance = None


def compiler():
    global _compiler_instance
    import lzma

    if _compiler_instance is not None:
        return _compiler_instance
    from qt.core import QApplication, QEventLoop, QFile, QIODevice, QObject, pyqtSlot
    from qt.webengine import QWebChannel, QWebEnginePage, QWebEngineScript

    from calibre.gui2 import must_use_qt
    from calibre.utils.webengine import secure_webengine, setup_default_profile, setup_profile

    must_use_qt()
    setup_default_profile()
    null = object()
    base = base_dir()
    rapydscript_dir = os.path.join(base, 'src', 'pyj')
    cache_path = os.path.join(module_cache_dir(), 'embedded-compiler-write-cache')
    os.makedirs(cache_path, exist_ok=True)

    class JSBridge(QObject):
        result: Any = null
        error: Any = null

        def __enter__(self):
            self.result = null
            self.error = null

        def __exit__(self, *a):
            pass

        def working(self):
            return self.result is null and self.error is null

        @pyqtSlot('QVariant')
        def on_result(self, data):
            self.result = data

        @pyqtSlot('QVariant')
        def on_error(self, data):
            self.error = data

        @pyqtSlot(str, str, bool)
        def write_file(self, name: str, data: str, is_binary: bool) -> None:
            if name.startswith('__vfs__/'):
                name = name.partition('/')[2]
                path = os.path.abspath(os.path.join(cache_path, name))
                if is_path_inside(cache_path, path):
                    os.makedirs(os.path.dirname(path), exist_ok=True)
                    if is_binary:
                        with open(path, 'wb') as fb:
                            fb.write(standard_b64decode(data))
                    else:
                        with open(path, 'w') as f:
                            f.write(data)

        @pyqtSlot(str, result='QVariant')
        def getmtime(self, name):
            if name.startswith('__vfs__/'):
                name = name.partition('/')[2]
                for base in (rapydscript_dir, cache_path):
                    path = os.path.abspath(os.path.join(base, name))
                    try:
                        return ['', int(os.path.getmtime(path) * 1000)]
                    except OSError:
                        pass
            return ['ENOENT', f'No file named {name} exists in {rapydscript_dir}']

        @pyqtSlot(str, str, result='QVariant')
        def read_file(self, name: str, encoding: str) -> list[str]:
            if name.startswith('__vfs__/'):
                name = name.partition('/')[2]
                for base in (rapydscript_dir, cache_path):
                    path = os.path.abspath(os.path.join(base, name))
                    if is_existing_subpath(path, base):
                        try:
                            with open(path, 'rb') as f:
                                ans = f.read()
                            if encoding:
                                payload = ans.decode(encoding)
                            else:
                                payload = standard_b64encode(ans).decode('ascii')
                            return ['', payload]
                        except OSError as e:
                            return ['EIO', str(e)]
            return ['ENOENT', f'No file named {name} exists in {rapydscript_dir}']

    with lzma.open(P(COMPILER_PATH, allow_user_override=False)) as lzf:
        compiler_script = lzf.read().decode('utf-8')

    def vfs_script():
        return '''
(function() {
"use strict";

async function stat_file(name) {
    const [a, b] = await py_bridge.getmtime(name);
    if (a.length === 0) return {mtimeMs: b};
    let e = new Error(b);
    e.code = a;
    throw e;
}

async function read_file(name, encoding) {
    const [a, b] = await py_bridge.read_file(name, encoding || '');
    if (a.length === 0) {
        if (encoding && encoding.length) return b;
        return Uint8Array.fromBase64(b);
    }
    let e = new Error(b);
    e.code = a;
    throw e;
}

if (typeof Uint8Array.prototype.toBase64 !== 'function') {
  Uint8Array.prototype.toBase64 = function (options = {}) {
    const alphabetType = options.alphabet || 'base64';
    const omitPadding = !!options.omitPadding;

    // 1. Standard Base64 Standard Alphabet vs Base64URL Alphabet
    const chars = alphabetType === 'base64url'
      ? 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'
      : 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';

    let result = '';
    const len = this.length;
    const extraBytes = len % 3;
    const mainLength = len - extraBytes;

    // 2. Process bytes in chunks of 3 (turns 3 bytes into 4 base64 characters)
    for (let i = 0; i < mainLength; i += 3) {
      const chunk = (this[i] << 16) | (this[i + 1] << 8) | this[i + 2];
      result += chars[(chunk >> 18) & 63] +
                chars[(chunk >> 12) & 63] +
                chars[(chunk >> 6) & 63] +
                chars[chunk & 63];
    }

    // 3. Handle remaining bytes (padding logic)
    if (extraBytes === 1) {
      const chunk = this[mainLength];
      result += chars[(chunk >> 2) & 63] + chars[(chunk << 4) & 63];
      if (!omitPadding) result += '==';
    } else if (extraBytes === 2) {
      const chunk = (this[mainLength] << 8) | this[mainLength + 1];
      result += chars[(chunk >> 10) & 63] + chars[(chunk >> 4) & 63] + chars[(chunk << 2) & 63];
      if (!omitPadding) result += '=';
    }

    return result;
  };
}

async function write_file(name, data) {
    let payload = data;
    let is_binary = false;
    if (data instanceof Uint8Array) {
        payload = data.toBase64();
        is_binary = true;
    } else if (typeof data !== 'string') {
        throw new Error("Unsupported data type!");
    }
    await py_bridge.write_file(name, payload, is_binary);
}

RapydScript.virtual_file_system = {
    read_file: read_file, write_file: write_file, stat_file: stat_file,
};

let compiler_lazy = {promise: RapydScript.create_embedded_compiler()};
async function compile(src, options) {
    if (!compiler_lazy.compiler) compiler_lazy.compiler = await compiler_lazy.promise;
    return await compiler_lazy.compiler.compile(src, options);
}
window.compile = compile;
new QWebChannel(qt.webChannelTransport, async function (channel) {
    window.py_bridge = channel.objects.py_bridge;
    document.title = 'compiler initialized';
})
})();
'''

    world = QWebEngineScript.ScriptWorldId.MainWorld

    def create_script(src, name):
        s = QWebEngineScript()
        s.setName(name)
        s.setInjectionPoint(QWebEngineScript.InjectionPoint.DocumentReady)
        s.setWorldId(world)
        s.setRunsOnSubFrames(True)
        s.setSourceCode(src)
        return s

    class Compiler(QWebEnginePage):
        def __init__(self):
            super().__init__()
            setup_profile(self.profile())
            secure_webengine(self, for_viewer=True)
            self.bridge = JSBridge()
            self.channel = QWebChannel()
            self.channel.registerObject('py_bridge', self.bridge)
            self.errors = []
            self.setWebChannel(self.channel)
            file = QFile(':/qtwebchannel/qwebchannel.js')
            if not file.open(QIODevice.OpenModeFlag.ReadOnly):
                raise RuntimeError('Could not load qwebchannel.js')
            qwc = bytes(file.readAll()).decode()
            script = compiler_script
            self.scripts().insert(create_script(qwc, 'qwebchannel.js'))
            self.scripts().insert(create_script(script, 'rapydscript.js'))
            self.scripts().insert(create_script(vfs_script(), 'vfs.js'))
            self.setHtml('<p>initialize')
            self.spin_loop(
                lambda: self.title() != 'compiler initialized',
                'Creating RapydScript compiler took too long',
                timeout=10,
            )

        def spin_loop(self, while_condition, timeout_err, timeout=60):
            limit = time.monotonic() + timeout
            while while_condition():
                app = QApplication.instance()
                assert app is not None
                app.processEvents(QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)
                if time.monotonic() > limit:
                    raise TimeoutError(timeout_err)

        def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
            print(f'{str(level).rpartition('.')[-1]}:{sourceID}:{lineNumber}:{message}', flush=True)

        def __call__(self, src, options) -> dict[str, str]:
            self.errors = []
            self.working = True
            options['write_name'] = True
            options['keep_docstrings'] = False
            src = '''
    (async () => {{
            try {{
                var js = await window.compile({}, {});
                py_bridge.on_result(js);
            }} catch (error) {{
                py_bridge.on_error({{name: error.name, message: error.message, stack: error.stack}})
            }}
    }})();
'''.format(*map(json.dumps, (src, options)))
            with self.bridge:
                self.runJavaScript(src, world)
                self.spin_loop(self.bridge.working, 'Compilation of RapydScript code took too long')
            if self.bridge.error is not null:
                e = self.bridge.error
                raise CompileFailure(f'Failed to compile rapydscript code with error: {e['name']}: {e['message']}\n{e['stack']}')
            return self.bridge.result

        def eval(self, js):
            eval_result = null

            def on_complete(result):
                nonlocal eval_result
                eval_result = result

            self.runJavaScript(js, world, on_complete)
            self.spin_loop(lambda: eval_result is null, 'eval of JS took too long')
            return eval_result

    _compiler_instance = Compiler()
    return _compiler_instance


class CompileFailure(ValueError):
    pass


_cache_dir = None


def module_cache_dir():
    global _cache_dir
    if _cache_dir is None:
        d = os.path.dirname
        base = d(d(d(d(abspath(__file__)))))
        _cache_dir = os.path.join(base, '.build-cache', 'pyj')
        try:
            os.makedirs(_cache_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    return _cache_dir


def ok_to_import_webengine():
    from qt.core import QApplication

    if QApplication.instance() is None:
        return True
    if 'PyQt6.QtWebEngineCore' in sys.modules:
        return True
    return False


OUTPUT_SENTINEL = b'-----RS webengine compiler output starts here------'


def forked_compile():
    c = compiler()
    stdin = getattr(sys.stdin, 'buffer', sys.stdin)
    raw = stdin.read()
    data = raw.decode('utf-8') if isinstance(raw, bytes) else raw
    options = json.loads(sys.argv[-1])
    result = c(data, options)
    stdout = getattr(sys.stdout, 'buffer', sys.stdout)
    stdout.write(OUTPUT_SENTINEL)  # type: ignore
    stdout.write(as_bytes(json.dumps(result)))  # type: ignore
    stdout.close()


def run_forked_compile(data, options) -> dict[str, str]:
    from calibre.debug import run_calibre_debug

    p = run_calibre_debug(
        '-c',
        'from calibre.utils.rapydscript import *; forked_compile()',
        json.dumps(options),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        headless=True,
    )
    stdout = p.communicate(as_bytes(data))[0]
    if p.wait() != 0:
        raise SystemExit(p.returncode)
    idx = stdout.find(OUTPUT_SENTINEL)
    return json.loads(stdout[idx + len(OUTPUT_SENTINEL) :])


def compile_pyj(
    data,
    filename='<stdin>',
    beautify=True,
    private_scope=True,
    libdir=None,
    omit_baselib=False,
    tree_shaking=True,
    source_map_line_offset=0,
) -> dict[str, str]:
    if isinstance(data, bytes):
        data = data.decode('utf-8')
    options = {
        'beautify': beautify,
        'private_scope': private_scope,
        'keep_baselib': not omit_baselib,
        'filename': filename,
        'source_map_line_offset': source_map_line_offset,
        'tree_shaking': tree_shaking,
        'source_map': 'embedded',
    }
    if not ok_to_import_webengine():
        result = run_forked_compile(data, options)
    else:
        try:
            c = compiler()
            result = c(data, options)
        except RuntimeError as err:
            if 'Cannot use Qt in non GUI thread' in str(err):
                result = run_forked_compile(data, options)
            else:
                raise
    return result


@lru_cache(maxsize=2)
def external_compiler_version() -> tuple[str, tuple[int, int, int]]:
    from calibre.utils.filenames import find_executable_in_path

    rs = find_executable_in_path('rapydscript')
    ver = (0, 0, 0)
    try:
        raw = subprocess.check_output([rs, '--version'])
    except Exception:
        return '', ver
    if raw.startswith(b'rapydscript-ng '):
        rver = raw.partition(b' ')[-1]
        try:
            qver = tuple(map(int, rver.split(b'.')))
            ver = qver[0], qver[1], qver[2]
        except Exception:
            return '', ver
        if ver < (0, 8, 5):
            return '', ver
    return rs, ver


def external_compiler() -> str:
    return external_compiler_version()[0]


def compile_fast(
    data,
    filename=None,
    beautify=True,
    private_scope=True,
    libdir=None,
    omit_baselib=False,
    tree_shaking=True,
    source_map_line_offset=0,
) -> dict[str, str]:
    if not (rs := external_compiler()):
        return compile_pyj(
            data,
            filename or '<stdin>',
            beautify,
            private_scope,
            libdir,
            omit_baselib,
            tree_shaking=tree_shaking,
            source_map_line_offset=source_map_line_offset,
        )
    args = ['--cache-dir', module_cache_dir()]
    if libdir:
        args += ['--import-path', libdir]
    if not beautify:
        args.append('--uglify')
    if not private_scope:
        args.append('--bare')
    if omit_baselib:
        args.append('--omit-baselib')
    if not isinstance(data, bytes):
        data = data.encode('utf-8')
    if filename:
        args.append('--filename-for-stdin'), args.append(filename)
    if tree_shaking:
        args.append('--tree-shaking')
    with tempfile.TemporaryDirectory() as tdir:
        smap = os.path.join(tdir, 'smap')
        js = os.path.join(tdir, 'js')
        args.extend(('--source-map', smap, '--source-map-line-offset', str(source_map_line_offset)))
        args.extend(('--output', js))
        p = subprocess.run([rs, 'compile'] + args, capture_output=True, input=data)
        if p.returncode != 0:
            raise CompileFailure(p.stderr.decode())
        with open(smap, encoding='utf-8') as sf, open(js, encoding='utf-8') as jf:
            return {'code': jf.read(), 'source_map': sf.read()}


def base_dir():
    d = os.path.dirname
    return d(d(d(d(os.path.abspath(__file__)))))


def atomic_write(base, name, content):
    name = os.path.join(base, name)
    tname = name + '.tmp'
    with open(tname, 'wb') as f:
        f.write(as_bytes(content))
    atomic_rename(tname, name)


def run_rapydscript_tests():
    from calibre.gui2 import must_use_qt

    must_use_qt()
    from calibre.utils.webengine import create_script, insert_scripts, secure_webengine, setup_default_profile, setup_fake_protocol, setup_profile

    setup_fake_protocol()
    setup_default_profile()
    from urllib.parse import parse_qs

    from qt.core import QApplication, QByteArray, QEventLoop, QUrl
    from qt.webengine import QWebEnginePage, QWebEngineProfile, QWebEngineScript, QWebEngineUrlRequestJob, QWebEngineUrlSchemeHandler

    from calibre.constants import FAKE_HOST, FAKE_PROTOCOL
    from calibre.gui2.viewer.web_view import send_reply

    base = base_dir()
    rapydscript_dir = os.path.join(base, 'src', 'pyj')
    fname = os.path.join(rapydscript_dir, 'test.pyj')
    with open(fname, 'rb') as f:
        result = compile_fast(f.read(), fname)

    class UrlSchemeHandler(QWebEngineUrlSchemeHandler):
        def __init__(self, parent=None):
            QWebEngineUrlSchemeHandler.__init__(self, parent)
            self.allowed_hosts = (FAKE_HOST,)
            self.registered_data = {}

        def requestStarted(self, a0):
            if bytes(a0.requestMethod()) != b'GET':
                return self.fail_request(a0, QWebEngineUrlRequestJob.Error.RequestDenied)
            url = a0.requestUrl()
            host = url.host()
            if host not in self.allowed_hosts:
                return self.fail_request(a0)
            q = parse_qs(url.query())
            if not q:
                return self.fail_request(a0)
            mt = q.get('mime-type', ('text/plain',))[0]
            data = q.get('data', ('',))[0].encode('utf-8')
            send_reply(a0, mt, data)

        def fail_request(self, rq, fail_code=None):
            if fail_code is None:
                fail_code = QWebEngineUrlRequestJob.Error.UrlNotFound
            rq.fail(fail_code)
            print(f'Blocking FAKE_PROTOCOL request: {rq.requestUrl().toString()}', file=sys.stderr)

    class Tester(QWebEnginePage):
        def __init__(self):
            profile = QWebEngineProfile(QApplication.instance())
            profile.setHttpUserAgent('calibre-tester')
            setup_profile(profile)
            insert_scripts(profile, create_script('test-rapydscript.js', result['code'], on_subframes=False))
            url_handler = UrlSchemeHandler(profile)
            profile.installUrlSchemeHandler(QByteArray(FAKE_PROTOCOL.encode('ascii')), url_handler)
            QWebEnginePage.__init__(self, profile, None)
            self.titleChanged.connect(self.title_changed)
            secure_webengine(self)
            self.setHtml('<p>initialize', QUrl(f'{FAKE_PROTOCOL}://{FAKE_HOST}/index.html'))
            self.working = True

        def title_changed(self, title):
            if title == 'initialized':
                self.titleChanged.disconnect()
                self.runJavaScript('window.main()', QWebEngineScript.ScriptWorldId.ApplicationWorld, self.callback)

        def spin_loop(self):
            while self.working:
                app = QApplication.instance()
                assert app is not None
                app.processEvents(QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)
            return self.result

        def callback(self, result):
            self.result = result
            self.working = False

        def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
            print(
                message,
                file=sys.stdout if level == QWebEnginePage.JavaScriptConsoleMessageLevel.InfoMessageLevel else sys.stderr,
            )

    tester = Tester()
    res = tester.spin_loop()
    if res is None:
        res = 1
    raise SystemExit(int(res))


def set_data(src, output: str, source_map: str = '', source_url: str = '', **kw) -> str:
    from calibre.db.constants import NO_SEARCH_LINK
    from calibre.ebooks.oeb.polish.main import SUPPORTED
    from calibre.library.page_count import CHARS_PER_PAGE

    for k, v in {
        '__SPECIAL_TITLE__': SPECIAL_TITLE_FOR_WEBENGINE_COMMS,
        '__FAKE_PROTOCOL__': FAKE_PROTOCOL,
        '__FAKE_HOST__': FAKE_HOST,
        '__CALIBRE_VERSION__': __version__,
        '__DARK_LINK_COLOR__': dark_link_color,
        '__BUILTIN_COLORS_LIGHT__': json.dumps(builtin_colors_light),
        '__BUILTIN_COLORS_DARK__': json.dumps(builtin_colors_dark),
        '__BUILTIN_DECORATIONS__': json.dumps(builtin_decorations),
        '__NO_SEARCH_LINK__': NO_SEARCH_LINK,
        '__CHARS_PER_PAGE__': str(CHARS_PER_PAGE),
        '__EDITABLE_FORMATS__': json.dumps(sorted(SUPPORTED)),
    }.items():
        src = src.replace(k, v, 1)
    for k, v in kw.items():
        src = src.replace(k, v, 1)
    src = src.rstrip()
    src = src.removesuffix('//# sourceMappingURL=smap')
    if source_map:
        from base64 import standard_b64encode

        payload = standard_b64encode(source_map.encode()).decode()
        smurl = f'//# sourceMappingURL=data:application/json;charset=utf-8;base64,{payload}'
    else:
        smurl = f'//# sourceMappingURL={output + '.map'}'
    if source_url:
        src += f'//# sourceURL={source_url}'
    return src + '\n' + smurl


def compile_editor():
    base = base_dir()
    rapydscript_dir = os.path.join(base, 'src', 'pyj')
    fname = os.path.join(rapydscript_dir, 'editor.pyj')
    base = os.path.join(base, 'resources')
    output = 'editor.js'
    with open(fname, 'rb') as f:
        result = compile_fast(f.read(), fname)
        js = set_data(result['code'], output)
    atomic_write(base, output, js)
    atomic_write(base, output + '.map', result['source_map'])


def compile_viewer():
    base = base_dir()
    iconf = os.path.join(base, 'imgsrc', 'srv', 'generate.py')
    g = {'__file__': iconf}
    exec_path(iconf, g)
    icons = g['merge']()
    with open(os.path.join(base, 'resources', 'content-server', 'reset.css'), 'rb') as f:
        reset = f.read().decode('utf-8')
    with open(os.path.join(base, 'resources', 'content-server', 'base.css'), 'rb') as f:
        base_css = f.read().decode('utf-8')
    html = f'<!DOCTYPE html>\n<html><head><style>{reset}</style><style>{base_css}</style></head><body>{icons}</body></html>'

    rapydscript_dir = os.path.join(base, 'src', 'pyj')
    fname = os.path.join(rapydscript_dir, 'viewer-main.pyj')
    output = 'viewer.js'
    with open(fname, 'rb') as f:
        result = compile_fast(f.read(), fname)
        js = set_data(result['code'], output)
    base = os.path.join(base, 'resources')
    atomic_write(base, output, js)
    atomic_write(base, output + '.map', result['source_map'])
    atomic_write(base, 'viewer.html', html)


def count_lines_before(text: str, qline: str) -> int:
    qline = qline.strip()
    for line_number, line in enumerate(io.StringIO(text)):
        if line.strip() == qline:
            return line_number
    raise ValueError(f'{qline} not found')


def compile_srv():
    base = base_dir()
    iconf = os.path.join(base, 'imgsrc', 'srv', 'generate.py')
    g = {'__file__': iconf}
    exec_path(iconf, g)
    icons = g['merge']()
    with open(os.path.join(base, 'resources', 'content-server', 'reset.css')) as f:
        reset = f.read()
    with open(os.path.join(base, 'resources', 'content-server', 'base.css')) as f:
        base_css = f.read()
    with open(os.path.join(base, 'src', 'pyj', 'book_list', 'constants.pyj')) as f:
        constants = f.read()
    m = re.search(r"^cs_top_bar_host_id = '(.+?)'", constants, flags=re.M)
    assert m is not None
    cs_top_bar_host_id = m.group(1)
    m = re.search(r"^book_list_container_id = '(.+?)'", constants, flags=re.M)
    assert m is not None
    book_list_container_id = m.group(1)
    m = re.search(r"^read_book_container_id = '(.+?)'", constants, flags=re.M)
    assert m is not None
    read_book_container_id = m.group(1)
    base_css = base_css.replace('CS_TOP_BAR_HOST_ID', cs_top_bar_host_id)
    base_css = base_css.replace('BOOK_LIST_CONTAINER_ID', book_list_container_id)
    base_css = base_css.replace('READ_BOOK_CONTAINER_ID', read_book_container_id)
    rapydscript_dir = os.path.join(base, 'src', 'pyj')
    rb = os.path.join(base, 'src', 'calibre', 'srv', 'render_book.py')
    with open(rb, 'rb') as f:
        rv_m = re.search(rb'^RENDER_VERSION\s+=\s+(\d+)', f.read(), re.M)
        assert rv_m is not None
        rv = str(int(rv_m.group(1)))
    mathjax_version = json.loads(P('mathjax/manifest.json', data=True, allow_user_override=False))['etag']
    base = os.path.join(base, 'resources', 'content-server')
    fname = os.path.join(rapydscript_dir, 'srv.pyj')
    output = 'index.js'
    with open(fname, 'rb') as f:
        result = compile_fast(f.read(), fname)
        js = set_data(result['code'], output, source_url=output, __RENDER_VERSION__=rv, __MATHJAX_VERSION__=mathjax_version)
    with open(os.path.join(base, 'index.html')) as f:
        html = f.read().replace('RESET_STYLES', reset, 1).replace('ICONS', icons, 1).replace('MAIN_JS', js, 1).replace('BASE_STYLES', base_css, 1)

    atomic_write(base, 'index-generated.html', html)
    atomic_write(base, output + '.map', result['source_map'])


def compile_all():
    compile_editor()
    compile_viewer()
    compile_srv()


# }}}

# Translations {{{


def create_pot(source_files):
    gettext_options = {
        'package_name': __appname__,
        'package_version': __version__,
        'bugs_address': 'https://bugs.launchpad.net/calibre',
    }
    if not (rs := external_compiler()):
        c = compiler()
        c.eval(f'window.catalog = {{}}; window.gettext_options = {json.dumps(gettext_options)}; 1')
        for fname in source_files:
            with open(fname, 'rb') as f:
                code = f.read().decode('utf-8')
            c.eval('RapydScript.gettext_parse(window.catalog, {}, {}); 1'.format(*map(json.dumps, (code, fname))))

        buf = c.eval('ans = []; RapydScript.gettext_output(window.catalog, window.gettext_options, ans.push.bind(ans)); ans;')
        return ''.join(buf)
    cp = subprocess.run(
        [
            rs,
            'gettext',
            '--package-name',
            gettext_options['package_name'],
            '--package-version',
            gettext_options['package_version'],
            '--bugs-address',
            gettext_options['bugs_address'],
        ]
        + list(source_files),
        capture_output=True,
    )
    if cp.returncode != 0:
        sys.stderr.buffer.write(cp.stderr)
        raise SystemExit(cp.returncode)
    return cp.stdout.decode().strip()


def msgfmt(po_data_as_string):
    if not (rs := external_compiler()):
        c = compiler()
        return c.eval('RapydScript.msgfmt({}, {})'.format(json.dumps(po_data_as_string), json.dumps({'use_fuzzy': False})))
    cp = subprocess.run([rs, 'msgfmt'], input=po_data_as_string.encode(), capture_output=True)
    if cp.returncode != 0:
        sys.stderr.write(cp.stderr)
        raise SystemExit(cp.returncode)
    return cp.stdout.decode().strip()


# }}}



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal kovid@kovidgoyal.net

"""
Read metadata from RAR archives
"""

import os
from io import BytesIO

from calibre.utils.unrar import extract_member, names


def get_metadata(stream):
    from calibre.ebooks.metadata.archive import is_comic
    from calibre.ebooks.metadata.meta import get_metadata

    file_names = list(names(stream))
    if is_comic(file_names):
        return get_metadata(stream, 'cbr')
    for f in file_names:
        stream_type = os.path.splitext(f)[1].lower()
        if stream_type:
            stream_type = stream_type[1:]
            if stream_type in {
                'lit',
                'opf',
                'prc',
                'mobi',
                'fb2',
                'epub',
                'rb',
                'imp',
                'pdf',
                'lrf',
                'azw',
                'azw1',
                'azw3',
            }:
                name, data = extract_member(stream, match=None, name=f)
                stream = BytesIO(data)
                stream.name = os.path.basename(name)
                mi = get_metadata(stream, stream_type)
                mi.timestamp = None
                return mi
    raise ValueError('No ebook found in RAR archive')
